import os
import sys
import argparse
import datetime
import ntpath
import torch
import torch.backends.cudnn as cudnn
import torch.distributed as dist
from shutil import copyfile
import numpy as np
import importlib
import multiprocessing

from utils.misc import init_devices, is_dist_avail_and_initialized, set_seed
from config.config import Config
from engine.check_configuration import check_configuration

if __name__ == '__main__':

    ##########################
    #   ARGS COMPROBATION    #
    ##########################

    # Normal exec:
    # python -u main.py \
    #     --config $input_job_cfg_file \        
    #     --result_dir $result_dir \        
    #     --name $job_name \          
    #     --run_id $job_counter \        
    #     --gpu 0
    # Distributed:
    # python -u -m torch.distributed.run \
    #     --nproc_per_node=2 \  
    #     main.py \
    #     --config $input_job_cfg_file \       
    #     --result_dir $result_dir \        
    #     --name $job_name \          
    #     --run_id $job_counter \        
    #     --gpu 0,1

    parser = argparse.ArgumentParser()
    parser.add_argument("--config", required=True, help="Path to the configuration file")
    parser.add_argument("--result_dir", help="Path to where the resulting output of the job will be stored",
                        default=os.getenv('HOME'))
    parser.add_argument("--name", help="Job name", default="unknown_job")
    parser.add_argument("--run_id", help="Run number of the same job", type=int, default=1)
    parser.add_argument("--gpu", help="GPU number according to 'nvidia-smi' command", type=str)

    # Distributed training parameters
    parser.add_argument('--world_size', default=1, type=int,
                        help='number of distributed processes')
    parser.add_argument('--local_rank', default=-1, type=int,
                        help='Node rank for distributed training. Necessary for using the torch.distributed.launch utility.')
    parser.add_argument('--dist_on_itp', action='store_true')
    parser.add_argument('--dist_url', default='env://',
                        help='url used to set up distributed training')
    parser.add_argument('--dist_backend', type=str, default='nccl', choices=['nccl', 'gloo'],
                        help='Backend to use in distributed mode')
    args = parser.parse_args()

    ############
    #  CHECKS  #
    ############

    # Job complete name
    job_identifier = args.name + '_' + str(args.run_id)

    # Prepare working dir
    job_dir = os.path.join(args.result_dir, args.name)
    cfg_bck_dir = os.path.join(job_dir, 'config_files')
    os.makedirs(cfg_bck_dir, exist_ok=True)
    head, tail = ntpath.split(args.config)
    cfg_filename = tail if tail else ntpath.basename(head)
    cfg_file = os.path.join(cfg_bck_dir,cfg_filename)

    if not os.path.exists(args.config):
        raise FileNotFoundError("Provided {} config file does not exist".format(args.config))
    copyfile(args.config, cfg_file)

    # Merge conf file with the default settings
    cfg = Config(job_dir, job_identifier)
    cfg._C.merge_from_file(cfg_file)
    cfg.update_dependencies()
    #cfg.freeze()

    now = datetime.datetime.now()
    print("Date: {}".format(now.strftime("%Y-%m-%d %H:%M:%S")))
    print("Arguments: {}".format(args))
    print("Job: {}".format(job_identifier))
    print("Python       : {}".format(sys.version.split('\n')[0]))
    print("PyTorch: ", torch.__version__)

    # GPU selection
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    opts = []
    if args.gpu:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu
        num_gpus = len(np.unique(np.array(args.gpu.strip().split(','))))
        opts.extend(["SYSTEM.NUM_GPUS", num_gpus])

    # GPU management
    device = init_devices(args, cfg.get_cfg_defaults())
    cfg._C.merge_from_list(opts)
    cfg = cfg.get_cfg_defaults()

    # Reproducibility
    set_seed(cfg.SYSTEM.SEED)
    
    # Number of CPU calculation
    if cfg.SYSTEM.NUM_CPUS == -1:
        cpu_count = multiprocessing.cpu_count()
    else:
        cpu_count = cfg.SYSTEM.NUM_CPUS
    torch.set_num_threads(cpu_count)
    cfg.merge_from_list(['SYSTEM.NUM_CPUS', cpu_count])

    check_configuration(cfg)
    print("Configuration details:")
    print(cfg)

    ##########################
    #       TRAIN/TEST       #
    ##########################
    workflowname = str(cfg.PROBLEM.TYPE).lower()
    mdl = importlib.import_module('engine.'+workflowname)
    names = [x for x in mdl.__dict__ if not x.startswith("_")]
    globals().update({k: getattr(mdl, k) for k in names})
    name = [x for x in names if "Base" not in x and "Workflow" in x][0]

    # Initialize workflow
    print("*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*")
    print(f"Initializing {name}")
    print("*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*\n")
    workflow = getattr(mdl, name)(cfg, job_identifier, device, args)

    if cfg.TRAIN.ENABLE:
        workflow.train()

    if cfg.TEST.ENABLE:
        workflow.test()

    if is_dist_avail_and_initialized():
        dist.barrier()
        dist.destroy_process_group()

    print("FINISHED JOB {} !!".format(job_identifier))




import torch
import numpy as np

from engine.base_workflow import Base_Workflow
from utils.util import save_tif, check_masks
from utils.misc import to_pytorch_format
from engine.metrics import jaccard_index, weighted_bce_dice_loss

class Semantic_Segmentation_Workflow(Base_Workflow):
    """
    Semantic segmentation workflow where the goal is to assign a class to each pixel of the input image. 
    More details in `our documentation <https://biapy.readthedocs.io/en/latest/workflows/semantic_segmentation.html>`_.  

    Parameters
    ----------
    cfg : YACS configuration
        Running configuration.
    
    Job_identifier : str
        Complete name of the running job.

    device : Torch device
        Device used. 

    args : argpase class
        Arguments used in BiaPy's call. 
    """
    def __init__(self, cfg, job_identifier, device, args, **kwargs):
        super(Semantic_Segmentation_Workflow, self).__init__(cfg, job_identifier, device, args, **kwargs)

        print("####################")      
        print("#  PRE-PROCESSING  #")
        print("####################")
        if cfg.TRAIN.ENABLE and cfg.DATA.TRAIN.CHECK_DATA:
            if cfg.LOSS.TYPE == 'MASKED_BCE':
                check_masks(cfg.DATA.TRAIN.GT_PATH, n_classes=3)
            else:
                check_masks(cfg.DATA.TRAIN.GT_PATH, n_classes=cfg.MODEL.N_CLASSES+1)
            if not cfg.DATA.VAL.FROM_TRAIN:
                if cfg.LOSS.TYPE == 'MASKED_BCE':
                    check_masks(cfg.DATA.VAL.GT_PATH, n_classes=3)
                else:
                    check_masks(cfg.DATA.VAL.GT_PATH, n_classes=cfg.MODEL.N_CLASSES+1)
        if cfg.TEST.ENABLE and cfg.DATA.TEST.LOAD_GT and cfg.DATA.TEST.CHECK_DATA:
            if cfg.LOSS.TYPE == 'MASKED_BCE':
                check_masks(cfg.DATA.TEST.GT_PATH, n_classes=3)
            else:
                check_masks(cfg.DATA.TEST.GT_PATH, n_classes=cfg.MODEL.N_CLASSES+1)

        # From now on, no modification of the cfg will be allowed
        self.cfg.freeze()

        # Activations for each output channel:
        # channel number : 'activation'
        self.activations = {':': 'CE_Sigmoid'}

        # Workflow specific training variables
        self.mask_path = cfg.DATA.TRAIN.GT_PATH
        self.load_Y_val = True

    def define_metrics(self):
        """
        Definition of self.metrics, self.metric_names and self.loss variables.
        """
        if self.cfg.LOSS.TYPE == "CE": 
            self.metrics = [jaccard_index]
            self.metric_names = ["jaccard_index"]
            if self.cfg.MODEL.N_CLASSES <= 2:
                self.loss = torch.nn.BCEWithLogitsLoss()
            else:
                self.loss = torch.nn.CrossEntropyLoss()
        elif self.cfg.LOSS.TYPE == "W_CE_DICE":
            self.metrics = [jaccard_index]
            self.metric_names = ["jaccard_index"]
            self.loss = weighted_bce_dice_loss(w_dice=0.66, w_bce=0.33)

    def metric_calculation(self, output, targets, metric_logger=None):
        """
        Execution of the metrics defined in :func:`~define_metrics` function. 

        Parameters
        ----------
        output : Torch Tensor
            Prediction of the model. 

        targets : Torch Tensor
            Ground truth to compare the prediction with. 

        metric_logger : MetricLogger, optional
            Class to be updated with the new metric(s) value(s) calculated. 
        
        Returns
        -------
        value : float
            Value of the metric for the given prediction. 
        """
        with torch.no_grad():
            train_iou = self.metrics[0](output, targets, self.device, num_classes=self.cfg.MODEL.N_CLASSES)
            train_iou = train_iou.item() if not torch.isnan(train_iou) else 0
            if metric_logger is not None:
                metric_logger.meters[self.metric_names[0]].update(train_iou)
            else:
                return train_iou

    def after_merge_patches(self, pred, filenames):
        """
        Steps need to be done after merging all predicted patches into the original image.

        Parameters
        ----------
        pred : Torch Tensor
            Model prediction.

        filenames : List of str
            Filenames of the predicted images.  
        """
        # Save simple binarization of predictions
        if pred.ndim == 4 and self.cfg.PROBLEM.NDIM == '3D':
            save_tif(np.expand_dims((pred>0.5).astype(np.uint8),0), self.cfg.PATHS.RESULT_DIR.PER_IMAGE_BIN,
                     filenames, verbose=self.cfg.TEST.VERBOSE)
        else:
            save_tif((pred>0.5).astype(np.uint8), self.cfg.PATHS.RESULT_DIR.PER_IMAGE_BIN, filenames,
                        verbose=self.cfg.TEST.VERBOSE)

    def after_merge_patches_by_chunks_proccess_patch(self, filename):
        """
        Place any code that needs to be done after merging all predicted patches into the original image
        but in the process made chunk by chunk. This function will operate patch by patch defined by 
        ``DATA.PATCH_SIZE``.

        Parameters
        ----------
        filename : List of str
            Filename of the predicted image H5/Zarr.  
        """
        pass

    def after_full_image(self, pred, filenames):
        """
        Steps that must be executed after generating the prediction by supplying the entire image to the model.

        Parameters
        ----------
        pred : Torch Tensor
            Model prediction.

        filenames : List of str
            Filenames of the predicted images.  
        """
        # Save simple binarization of predictions
        if pred.ndim == 4 and self.cfg.PROBLEM.NDIM == '3D':
            save_tif(np.expand_dims((pred>0.5).astype(np.uint8),0), self.cfg.PATHS.RESULT_DIR.FULL_IMAGE_BIN,
                     filenames, verbose=self.cfg.TEST.VERBOSE)
        else:
            save_tif((pred>0.5).astype(np.uint8), self.cfg.PATHS.RESULT_DIR.FULL_IMAGE_BIN, filenames,
                        verbose=self.cfg.TEST.VERBOSE)

    def after_all_images(self):
        """
        Steps that must be done after predicting all images. 
        """
        super().after_all_images()

    def normalize_stats(self, image_counter):
        """
        Normalize statistics.  

        Parameters
        ----------
        image_counter : int
            Number of images to average the metrics.
        """
        super().normalize_stats(image_counter)

    def print_stats(self, image_counter):
        """
        Print statistics.  

        Parameters
        ----------
        image_counter : int
            Number of images to call ``normalize_stats``.
        """
        super().print_stats(image_counter)
        super().print_post_processing_stats()


        


import os
import torch
import numpy as np
import pandas as pd
from skimage.io import imread
from tqdm import tqdm
from skimage.segmentation import clear_border

from data.post_processing.post_processing import (watershed_by_channels, voronoi_on_mask, 
    remove_by_properties, repare_large_blobs)
from data.pre_processing import create_instance_channels, create_test_instance_channels
from utils.util import save_tif
from utils.matching import matching, wrapper_matching_dataset_lazy
from engine.metrics import jaccard_index, instance_segmentation_loss
from engine.base_workflow import Base_Workflow

class Instance_Segmentation_Workflow(Base_Workflow):
    """
    Instance segmentation workflow where the goal is to assign an unique id, i.e. integer, to each object of the input image.
    More details in `our documentation <https://biapy.readthedocs.io/en/latest/workflows/instance_segmentation.html>`_.  

    Parameters
    ----------
    cfg : YACS configuration
        Running configuration.
    
    Job_identifier : str
        Complete name of the running job.

    device : Torch device
        Device used. 

    args : argpase class
        Arguments used in BiaPy's call. 
    """
    def __init__(self, cfg, job_identifier, device, args, **kwargs):
        super(Instance_Segmentation_Workflow, self).__init__(cfg, job_identifier, device, args, **kwargs)

        self.original_test_path, self.original_test_mask_path = self.prepare_instance_data()

        self.all_matching_stats = []

        self.instance_ths = {}
        self.instance_ths['TYPE'] = self.cfg.PROBLEM.INSTANCE_SEG.DATA_MW_TH_TYPE
        self.instance_ths['TH_BINARY_MASK'] = self.cfg.PROBLEM.INSTANCE_SEG.DATA_MW_TH_BINARY_MASK
        self.instance_ths['TH_CONTOUR'] = self.cfg.PROBLEM.INSTANCE_SEG.DATA_MW_TH_CONTOUR
        self.instance_ths['TH_FOREGROUND'] = self.cfg.PROBLEM.INSTANCE_SEG.DATA_MW_TH_FOREGROUND
        self.instance_ths['TH_DISTANCE'] = self.cfg.PROBLEM.INSTANCE_SEG.DATA_MW_TH_DISTANCE
        self.instance_ths['TH_POINTS'] = self.cfg.PROBLEM.INSTANCE_SEG.DATA_MW_TH_POINTS 

        # From now on, no modification of the cfg will be allowed
        self.cfg.freeze()

        # Activations for each output channel:
        # channel number : 'activation'
        if self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS == "Dv2":
            self.activations = {'0': 'Linear'}
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS in ["BC", "BP", "BCM"]:
            self.activations = {':': 'CE_Sigmoid'}
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS in ["BDv2", "BD"]:
            self.activations = {'0': 'CE_Sigmoid', '1': 'Linear'}
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS in ["BCD", "BCDv2"]:
            self.activations = {'0': 'CE_Sigmoid', '1': 'CE_Sigmoid', '2': 'Linear'}

        # Workflow specific training variables
        self.mask_path = cfg.DATA.TRAIN.GT_PATH
        self.load_Y_val = True

        # Specific instance segmentation post-processing
        self.remove_by_prop = False 
        for i in range(len(self.cfg.TEST.POST_PROCESSING.REMOVE_BY_PROPERTIES)):
            if len(self.cfg.TEST.POST_PROCESSING.REMOVE_BY_PROPERTIES[i]) > 0:
                self.remove_by_prop = True
                break 
        if self.cfg.TEST.POST_PROCESSING.VORONOI_ON_MASK or self.remove_by_prop or\
            self.cfg.TEST.POST_PROCESSING.REPARE_LARGE_BLOBS_SIZE != -1:
            self.post_processing['instance_post'] = True
            self.all_matching_stats_post_processing = []   
        else:
            self.post_processing['instance_post'] = False            

    def define_metrics(self):
        """
        Definition of self.metrics, self.metric_names and self.loss variables.
        """
        self.metrics = []
        self.metric_names = []
        self.loss = instance_segmentation_loss(self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNEL_WEIGHTS,
            self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, 
            self.cfg.PROBLEM.INSTANCE_SEG.DISTANCE_CHANNEL_MASK)
        if self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS in ["BC", "BCM", "BP"]:
            self.metrics.append(jaccard_index)
            self.first_not_binary_channel = len(self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS)
            self.metric_names.append("jaccard_index")
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS == "BD":
            self.metrics.append(jaccard_index)
            self.first_not_binary_channel = 1 
            self.metric_names.append("jaccard_index")
            self.metrics.append(torch.nn.L1Loss())
            self.metric_names.append("L1_distance_channel")
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS in ["BCD", 'BCDv2']:
            self.metrics.append(jaccard_index)
            self.first_not_binary_channel = 2
            self.metric_names.append("jaccard_index")
            self.metrics.append(torch.nn.L1Loss())
            self.metric_names.append("L1_distance_channel")
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS == "BDv2":
            self.metrics.append(jaccard_index)
            self.first_not_binary_channel = 1
            self.metrics.append(torch.nn.L1Loss())
            self.metric_names.append("jaccard_index")
            self.metric_names.append("L1_distance_channel")
        elif self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS == "Dv2":
            self.metrics.append(torch.nn.L1Loss())
            self.metric_names.append("L1_distance_channel")

    def metric_calculation(self, output, targets, metric_logger=None):
        """
        Execution of the metrics defined in :func:`~define_metrics` function. 

        Parameters
        ----------
        output : Torch Tensor
            Prediction of the model. 

        targets : Torch Tensor
            Ground truth to compare the prediction with. 

        metric_logger : MetricLogger, optional
            Class to be updated with the new metric(s) value(s) calculated. 
        
        Returns
        -------
        value : float
            Value of the metric for the given prediction. 
        """
        with torch.no_grad():
            train_iou = []
            train_dis = []
            for i in range(len(self.metric_names)):
                if self.metric_names[i] == "jaccard_index":
                    iou = self.metrics[i](output, targets, self.device, num_classes=1, 
                        first_not_binary_channel=self.first_not_binary_channel)
                    iou = iou.item() if not torch.isnan(iou) else 0
                    train_iou.append(iou)
                elif self.metric_names[i] == "L1_distance_channel":
                    train_dis.append(self.metrics[i](output, targets).item())
            train_iou = np.mean(train_iou)
            train_dis = np.mean(train_dis)
            if metric_logger is not None:
                metric_logger.meters[self.metric_names[0]].update(train_iou)
                if len(self.metric_names) > 1:
                    metric_logger.meters[self.metric_names[1]].update(train_dis)
            else:
                return train_iou

    def instance_seg_process(self, pred, filenames):
        """
        Instance segmentation workflow engine for test/inference. Process model's prediction to prepare 
        instance segmentation output and calculate metrics. 

        Parameters
        ----------
        pred : Torch Tensor
            Model predictions.
        
        filenames : List of str
            Predicted image's filenames.
        """
        #############################
        ### INSTANCE SEGMENTATION ###
        #############################
        print("Creating instances with watershed . . .")
        w_dir = os.path.join(self.cfg.PATHS.WATERSHED_DIR, filenames[0])
        check_wa = w_dir if self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHECK_MW else None
        
        w_pred = watershed_by_channels(pred, self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, ths=self.instance_ths, 
            remove_before=self.cfg.PROBLEM.INSTANCE_SEG.DATA_REMOVE_BEFORE_MW, thres_small_before=self.cfg.PROBLEM.INSTANCE_SEG.DATA_REMOVE_SMALL_OBJ_BEFORE,
            seed_morph_sequence=self.cfg.PROBLEM.INSTANCE_SEG.SEED_MORPH_SEQUENCE, seed_morph_radius=self.cfg.PROBLEM.INSTANCE_SEG.SEED_MORPH_RADIUS, 
            erode_and_dilate_foreground=self.cfg.PROBLEM.INSTANCE_SEG.ERODE_AND_DILATE_FOREGROUND, fore_erosion_radius=self.cfg.PROBLEM.INSTANCE_SEG.FORE_EROSION_RADIUS, 
            fore_dilation_radius=self.cfg.PROBLEM.INSTANCE_SEG.FORE_DILATION_RADIUS, rmv_close_points=self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS, 
            remove_close_points_radius=self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS_RADIUS[0], resolution=self.cfg.DATA.TEST.RESOLUTION, save_dir=check_wa)

        save_tif(np.expand_dims(np.expand_dims(w_pred,-1),0), self.cfg.PATHS.RESULT_DIR.PER_IMAGE_INSTANCES,
            filenames, verbose=self.cfg.TEST.VERBOSE)

        # Add extra dimension if working in 2D
        if w_pred.ndim == 2:
            w_pred = np.expand_dims(w_pred,0)

        if self.cfg.TEST.MATCHING_STATS and (self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST):
            print("Calculating matching stats . . .")

            # Need to load instance labels, as Y are binary channels used for IoU calculation
            if self.cfg.TEST.ANALIZE_2D_IMGS_AS_3D_STACK:
                del self.Y
                _Y = np.zeros(w_pred.shape, dtype=w_pred.dtype)
                for i in range(len(self.test_filenames)):
                    test_file = os.path.join(self.original_test_mask_path, self.test_filenames[i])
                    _Y[i] = imread(test_file).squeeze()
            else:
                test_file = os.path.join(self.original_test_mask_path, self.test_filenames[self.f_numbers[0]])
                _Y = imread(test_file).squeeze()

            if _Y.ndim == 2: _Y = np.expand_dims(_Y,0)

            # Convert instances to integer
            if _Y.dtype == np.float32: _Y = _Y.astype(np.uint32)
            if _Y.dtype == np.float64: _Y = _Y.astype(np.uint64)

            diff_ths_colored_img = abs(len(self.cfg.TEST.MATCHING_STATS_THS_COLORED_IMG) - len(self.cfg.TEST.MATCHING_STATS_THS))
            colored_img_ths = self.cfg.TEST.MATCHING_STATS_THS_COLORED_IMG+[-1]*diff_ths_colored_img

            results = matching(_Y, w_pred, thresh=self.cfg.TEST.MATCHING_STATS_THS, report_matches=True)
            for i in range(len(results)):
                # Extract TPs, FPs and FNs from the resulting matching data structure 
                r_stats = results[i] 
                thr = r_stats['thresh']

                # TP and FN
                gt_ids = r_stats['gt_ids'][1:]
                matched_pairs = r_stats['matched_pairs']
                gt_match = [x[0] for x in matched_pairs]
                gt_unmatch = [x for x in gt_ids if x not in gt_match]
                matched_scores = list(r_stats['matched_scores'])+[0 for _ in gt_unmatch]
                pred_match = [x[1] for x in matched_pairs]+[-1 for _ in gt_unmatch]
                tag = ["TP" if score >= thr else "FN" for score in matched_scores]

                # FPs
                pred_ids = r_stats['pred_ids'][1:]
                fp_instances = [x for x in pred_ids if x not in pred_match]
                fp_instances += [pred_id for score, pred_id in zip(matched_scores, pred_match) if score < thr]

                # Save csv files
                df = pd.DataFrame(zip(gt_match+gt_unmatch, pred_match, matched_scores, tag), columns =['gt_id', 'pred_id', 'iou', 'tag'])
                df = df.sort_values(by=['gt_id'])  
                df_fp = pd.DataFrame(zip(fp_instances), columns =['pred_id'])

                os.makedirs(self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS, exist_ok=True)
                df.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS, os.path.splitext(filenames[0])[0]+'_th_{}_gt_assoc.csv'.format(thr)), index=False)
                df_fp.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS, os.path.splitext(filenames[0])[0]+'_th_{}_fp.csv'.format(thr)), index=False)
                del r_stats['matched_scores']; del r_stats['matched_tps']; del r_stats['matched_pairs']; del r_stats['pred_ids']; del r_stats['gt_ids']
                print("DatasetMatching: {}".format(r_stats))

                if colored_img_ths[i] != -1 and colored_img_ths[i] == thr:
                    print("Creating the image with a summary of detected points and false positives with colors . . .")
                    colored_result = np.zeros(w_pred.shape+(3,), dtype=np.uint8)

                    print("Painting TPs and FNs . . .")
                    for j in tqdm(range(len(gt_match))):
                        color = (0,255,0) if tag[j] == "TP" else (255,0,0) # Green or red
                        colored_result[np.where(_Y == gt_match[j])] = color
                    for j in tqdm(range(len(gt_unmatch))):
                        colored_result[np.where(_Y == gt_unmatch[j])] = (255,0,0) # Red

                    print("Painting FPs . . .")
                    for j in tqdm(range(len(fp_instances))):
                        colored_result[np.where(w_pred == fp_instances[j])] = (0,0,255) # Blue

                    save_tif(np.expand_dims(colored_result,0), self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS,
                            [os.path.splitext(filenames[0])[0]+'_th_{}.tif'.format(thr)], verbose=self.cfg.TEST.VERBOSE)          
                    del colored_result
            self.all_matching_stats.append(results)


        ###################
        # Post-processing #
        ###################
        if self.cfg.TEST.POST_PROCESSING.REPARE_LARGE_BLOBS_SIZE != -1:
            w_pred = repare_large_blobs(w_pred, self.cfg.TEST.POST_PROCESSING.REPARE_LARGE_BLOBS_SIZE)

        if self.remove_by_prop:
            w_pred, labels, npixels, areas, circularities, diameters, comment, all_conds = remove_by_properties(w_pred, self.cfg.DATA.TEST.RESOLUTION, 
                properties=self.cfg.TEST.POST_PROCESSING.REMOVE_BY_PROPERTIES, prop_values=self.cfg.TEST.POST_PROCESSING.REMOVE_BY_PROPERTIES_VALUES,
                comp_signs=self.cfg.TEST.POST_PROCESSING.REMOVE_BY_PROPERTIES_SIGN)

            # Save all instance stats
            size_measure = 'area' if w_pred.ndim == 2 else 'volume'
            df = pd.DataFrame(zip(np.array(labels, dtype=np.uint64), npixels, areas, circularities, diameters, comment, all_conds),  
                columns=['label','npixels', size_measure, 'circularity', 'diameter', 'comment', 'conditions'])
            df = df.sort_values(by=['label'])   
            df.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.PER_IMAGE_INSTANCES, os.path.splitext(filenames[0])[0]+'_full_stats.csv'), index=False)
            # Save only remain instances stats
            df = df[df["comment"].str.contains("Strange")==False] 
            os.makedirs(self.cfg.PATHS.RESULT_DIR.PER_IMAGE_POST_PROCESSING, exist_ok=True)
            df.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.PER_IMAGE_POST_PROCESSING, os.path.splitext(filenames[0])[0]+'_filtered_stats.csv'), index=False)
            del df

        if self.cfg.TEST.POST_PROCESSING.VORONOI_ON_MASK:
            w_pred = voronoi_on_mask(w_pred, pred, th=self.cfg.TEST.POST_PROCESSING.VORONOI_TH, verbose=self.cfg.TEST.VERBOSE)
            del pred

        if self.cfg.TEST.POST_PROCESSING.CLEAR_BORDER:
            print("Clearing borders . . .")
            w_pred = clear_border(w_pred)

        if self.post_processing['instance_post']:
            save_tif(np.expand_dims(np.expand_dims(w_pred,-1),0), self.cfg.PATHS.RESULT_DIR.PER_IMAGE_POST_PROCESSING,
                filenames, verbose=self.cfg.TEST.VERBOSE)

            if self.cfg.TEST.MATCHING_STATS and (self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST):
                print("Calculating matching stats after post-processing . . .")
                results = matching(_Y, w_pred, thresh=self.cfg.TEST.MATCHING_STATS_THS, report_matches=True)
                
                for i in range(len(results)):
                    # Extract TPs, FPs and FNs from the resulting matching data structure 
                    r_stats = results[i] 
                    thr = r_stats['thresh']

                    # TP and FN
                    gt_ids = r_stats['gt_ids'][1:]
                    matched_pairs = r_stats['matched_pairs']
                    gt_match = [x[0] for x in matched_pairs]
                    gt_unmatch = [x for x in gt_ids if x not in gt_match]
                    matched_scores = list(r_stats['matched_scores'])+[0 for _ in gt_unmatch]
                    pred_match = [x[1] for x in matched_pairs]+[-1 for _ in gt_unmatch]
                    tag = ["TP" if score >= thr else "FN" for score in matched_scores]

                    # FPs
                    pred_ids = r_stats['pred_ids'][1:]
                    fp_instances = [x for x in pred_ids if x not in pred_match]
                    fp_instances += [pred_id for score, pred_id in zip(matched_scores, pred_match) if score < thr]

                    # Save csv files
                    df = pd.DataFrame(zip(gt_match+gt_unmatch, pred_match, matched_scores, tag), columns =['gt_id', 'pred_id', 'iou', 'tag'])
                    df = df.sort_values(by=['gt_id'])  
                    df_fp = pd.DataFrame(zip(fp_instances), columns =['pred_id'])

                    os.makedirs(self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS, exist_ok=True)
                    df.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS, os.path.splitext(filenames[0])[0]+'_post-proc_th_{}_gt_assoc.csv'.format(thr)), index=False)
                    df_fp.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS, os.path.splitext(filenames[0])[0]+'_post-proc_th_{}_fp.csv'.format(thr)), index=False)
                    del r_stats['matched_scores']; del r_stats['matched_tps']; del r_stats['matched_pairs']; del r_stats['pred_ids']; del r_stats['gt_ids']
                    print("DatasetMatching: {}".format(r_stats))

                    if colored_img_ths[i] != -1 and colored_img_ths[i] == thr:
                        print("Creating the image with a summary of detected points and false positives with colors . . .")
                        colored_result = np.zeros(w_pred.shape+(3,), dtype=np.uint8)

                        print("Painting TPs and FNs . . .")
                        for j in tqdm(range(len(gt_match))):
                            color = (0,255,0) if tag[j] == "TP" else (255,0,0) # Green or red
                            colored_result[np.where(_Y == gt_match[j])] = color
                        for j in tqdm(range(len(gt_unmatch))):
                            colored_result[np.where(_Y == gt_unmatch[j])] = (255,0,0) # Red
                            
                        print("Painting FPs . . .")
                        for j in tqdm(range(len(fp_instances))):
                            colored_result[np.where(w_pred == fp_instances[j])] = (0,0,255) # Blue

                        save_tif(np.expand_dims(colored_result,0), self.cfg.PATHS.RESULT_DIR.INST_ASSOC_POINTS,
                                [os.path.splitext(filenames[0])[0]+'_post-proc_th_{}.tif'.format(thr)], verbose=self.cfg.TEST.VERBOSE)          
                        del colored_result
                self.all_matching_stats_post_processing.append(results)

    def after_merge_patches(self, pred, filenames):
        """
        Steps need to be done after merging all predicted patches into the original image.

        Parameters
        ----------
        pred : Torch Tensor
            Model prediction.

        filenames : List of str
            Filenames of the predicted images.  
        """
        if not self.cfg.TEST.ANALIZE_2D_IMGS_AS_3D_STACK:
            self.instance_seg_process(pred, filenames)        

    def after_merge_patches_by_chunks_proccess_patch(self, filename):
        """
        Place any code that needs to be done after merging all predicted patches into the original image
        but in the process made chunk by chunk. This function will operate patch by patch defined by 
        ``DATA.PATCH_SIZE``.

        Parameters
        ----------
        filename : List of str
            Filename of the predicted image H5/Zarr.  
        """
        pass

    def after_full_image(self, pred, filenames):
        """
        Steps that must be executed after generating the prediction by supplying the entire image to the model.

        Parameters
        ----------
        pred : Torch Tensor
            Model prediction.

        filenames : List of str
            Filenames of the predicted images.  
        """
        pass

    def after_all_images(self):
        """
        Steps that must be done after predicting all images. 
        """
        super().after_all_images()
        if self.cfg.TEST.ANALIZE_2D_IMGS_AS_3D_STACK:
            print("Analysing all images as a 3D stack . . .")    
            if type(self.all_pred) is list:
                self.all_pred = np.concatenate(self.all_pred)
                self.all_gt = np.concatenate(self.all_gt) if (self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST) else None
            self.instance_seg_process(self.all_pred,  self.all_gt, ["3D_stack.tif"], [])

    def normalize_stats(self, image_counter): 
        """
        Normalize statistics.  

        Parameters
        ----------
        image_counter : int
            Number of images to average the metrics.
        """
        super().normalize_stats(image_counter) 

        if (self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST):
            if self.cfg.TEST.MATCHING_STATS:
                self.stats['inst_stats'] = wrapper_matching_dataset_lazy(self.all_matching_stats, self.cfg.TEST.MATCHING_STATS_THS)
                if self.post_processing['instance_post']:
                    self.stats['inst_stats_vor'] = wrapper_matching_dataset_lazy(self.all_matching_stats_post_processing, self.cfg.TEST.MATCHING_STATS_THS)

    def print_stats(self, image_counter):
        """
        Print statistics.  

        Parameters
        ----------
        image_counter : int
            Number of images to call ``normalize_stats``.
        """
        super().print_stats(image_counter)
        super().print_post_processing_stats()

        print("Instance segmentation specific metrics:")
        if self.cfg.TEST.MATCHING_STATS and (self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST) :
            for i in range(len(self.cfg.TEST.MATCHING_STATS_THS)):
                print("IoU TH={}".format(self.cfg.TEST.MATCHING_STATS_THS[i]))
                print(self.stats['inst_stats'][i])
                if self.post_processing['instance_post']:
                    print("IoU (post-processing) TH={}".format(self.cfg.TEST.MATCHING_STATS_THS[i]))
                    print(self.stats['inst_stats_vor'][i])

    def prepare_instance_data(self):
        """
        Creates instance segmentation ground truth images to train the model based on the ground truth instances provided.
        They will be saved in a separate folder in the root path of the ground truth. 
        """
        print("###########################")
        print("#  PREPARE INSTANCE DATA  #")
        print("###########################")
        original_test_path, original_test_mask_path = None, None

        # Create selected channels for train data
        if (self.cfg.TRAIN.ENABLE or self.cfg.DATA.TEST.USE_VAL_AS_TEST) and (not os.path.isdir(self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_DIR) or \
            not os.path.isdir(self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_MASK_DIR)):
            print("You select to create {} channels from given instance labels and no file is detected in {}. "
                    "So let's prepare the data. Notice that, if you do not modify 'DATA.TRAIN.INSTANCE_CHANNELS_DIR' "
                    "path, this process will be done just once!".format(self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS,
                    self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_DIR))
            create_instance_channels(self.cfg)

        # Create selected channels for val data
        if self.cfg.TRAIN.ENABLE and not self.cfg.DATA.VAL.FROM_TRAIN and (not os.path.isdir(self.cfg.DATA.VAL.INSTANCE_CHANNELS_DIR) or \
            not os.path.isdir(self.cfg.DATA.VAL.INSTANCE_CHANNELS_MASK_DIR)):
            print("You select to create {} channels from given instance labels and no file is detected in {}. "
                    "So let's prepare the data. Notice that, if you do not modify 'DATA.VAL.INSTANCE_CHANNELS_DIR' "
                    "path, this process will be done just once!".format(self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS,
                    self.cfg.DATA.VAL.INSTANCE_CHANNELS_DIR))
            create_instance_channels(self.cfg, data_type='val')

        # Create selected channels for test data once
        if self.cfg.TEST.ENABLE and not self.cfg.DATA.TEST.USE_VAL_AS_TEST and (not os.path.isdir(self.cfg.DATA.TEST.INSTANCE_CHANNELS_DIR) or \
            (not os.path.isdir(self.cfg.DATA.TEST.INSTANCE_CHANNELS_MASK_DIR) and self.cfg.DATA.TEST.LOAD_GT)) and not self.cfg.TEST.BY_CHUNKS:
            print("You select to create {} channels from given instance labels and no file is detected in {}. "
                    "So let's prepare the data. Notice that, if you do not modify 'DATA.TEST.INSTANCE_CHANNELS_DIR' "
                    "path, this process will be done just once!".format(self.cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS,
                    self.cfg.DATA.TEST.INSTANCE_CHANNELS_MASK_DIR))
            create_test_instance_channels(self.cfg)

        opts = []
        if self.cfg.TRAIN.ENABLE:
            print("DATA.TRAIN.PATH changed from {} to {}".format(self.cfg.DATA.TRAIN.PATH, self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_DIR))
            print("DATA.TRAIN.GT_PATH changed from {} to {}".format(self.cfg.DATA.TRAIN.GT_PATH, self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_MASK_DIR))
            opts.extend(['DATA.TRAIN.PATH', self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_DIR,
                        'DATA.TRAIN.GT_PATH', self.cfg.DATA.TRAIN.INSTANCE_CHANNELS_MASK_DIR])
        if not self.cfg.DATA.VAL.FROM_TRAIN:
            print("DATA.VAL.PATH changed from {} to {}".format(self.cfg.DATA.VAL.PATH, self.cfg.DATA.VAL.INSTANCE_CHANNELS_DIR))
            print("DATA.VAL.GT_PATH changed from {} to {}".format(self.cfg.DATA.VAL.GT_PATH, self.cfg.DATA.VAL.INSTANCE_CHANNELS_MASK_DIR))
            opts.extend(['DATA.VAL.PATH', self.cfg.DATA.VAL.INSTANCE_CHANNELS_DIR,
                        'DATA.VAL.GT_PATH', self.cfg.DATA.VAL.INSTANCE_CHANNELS_MASK_DIR])
        if self.cfg.TEST.ENABLE and not self.cfg.DATA.TEST.USE_VAL_AS_TEST and not self.cfg.TEST.BY_CHUNKS:
            print("DATA.TEST.PATH changed from {} to {}".format(self.cfg.DATA.TEST.PATH, self.cfg.DATA.TEST.INSTANCE_CHANNELS_DIR))
            opts.extend(['DATA.TEST.PATH', self.cfg.DATA.TEST.INSTANCE_CHANNELS_DIR])
            if self.cfg.DATA.TEST.LOAD_GT:
                print("DATA.TEST.GT_PATH changed from {} to {}".format(self.cfg.DATA.TEST.GT_PATH, self.cfg.DATA.TEST.INSTANCE_CHANNELS_MASK_DIR))
                opts.extend(['DATA.TEST.GT_PATH', self.cfg.DATA.TEST.INSTANCE_CHANNELS_MASK_DIR])
        original_test_path = self.cfg.DATA.TEST.PATH
        original_test_mask_path = self.cfg.DATA.TEST.GT_PATH
        self.cfg.merge_from_list(opts)

        return original_test_path, original_test_mask_path




import torch
import numpy as np
import torch.nn as nn
from timm.layers.format import Format
from typing import Callable, List, Optional, Tuple

class PatchEmbed(nn.Module):
    """ 2D Image to Patch Embedding
    """
    output_fmt: Format

    def __init__(
            self,
            img_size: Optional[int] = 224,
            patch_size: int = 16,
            in_chans: int = 3,
            ndim: int = 2, 
            embed_dim: int = 768,
            norm_layer: Optional[Callable] = None,
            flatten: bool = True,
            output_fmt: Optional[str] = None,
            bias: bool = True,
            strict_img_size: bool = True,
    ):
        super().__init__()
        self.ndim = ndim
        self.patch_size = (patch_size,)*self.ndim
        if img_size is not None:
            self.img_size = (img_size,)*self.ndim
            self.grid_size = tuple([s // p for s, p in zip(self.img_size, self.patch_size)])
            self.num_patches = self.grid_size[0]**self.ndim
        else:
            self.img_size = None
            self.grid_size = None
            self.num_patches = None

        if output_fmt is not None:
            self.flatten = False
            self.output_fmt = Format(output_fmt)
        else:
            # flatten spatial dim and transpose to channels last, kept for bwd compat
            self.flatten = flatten
            self.output_fmt = Format.NCHW
        self.strict_img_size = strict_img_size

        if self.ndim == 2:
            self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size, bias=bias)
        else:
            self.proj = nn.Conv3d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size, bias=bias)
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()
        
    def forward(self, x):
        if self.ndim == 2:
            B, C, H, W = x.shape
        else:
            B, C, Z, H, W = x.shape
        if self.img_size is not None:
            if self.strict_img_size:
                if self.ndim == 2:
                    assert H == self.img_size[0], f"Input height ({H}) doesn't match model ({self.img_size[0]})."
                    assert W == self.img_size[1], f"Input width ({W}) doesn't match model ({self.img_size[1]})."
                else:
                    assert Z == self.img_size[0], f"Input depth ({Z}) doesn't match model ({self.img_size[0]})."
                    assert H == self.img_size[1], f"Input height ({H}) doesn't match model ({self.img_size[1]})."
                    assert W == self.img_size[2], f"Input width ({W}) doesn't match model ({self.img_size[2]})."
            else:
                if self.ndim == 2:
                    assert Z % self.patch_size[0] == 0, f"Input height ({Z}) should be divisible by patch size ({self.patch_size[0]})."
                    assert H % self.patch_size[1] == 0, f"Input height ({H}) should be divisible by patch size ({self.patch_size[1]})."
                    assert W % self.patch_size[2] == 0, f"Input width ({W}) should be divisible by patch size ({self.patch_size[2]})."
                else:
                    assert H % self.patch_size[0] == 0, f"Input height ({H}) should be divisible by patch size ({self.patch_size[0]})."
                    assert W % self.patch_size[1] == 0, f"Input width ({W}) should be divisible by patch size ({self.patch_size[1]})."

        x = self.proj(x)
        if self.flatten:
            x = x.flatten(2).transpose(1, 2)  # NCHW -> NLC
        elif self.output_fmt != Format.NCHW:
            x = nchw_to(x, self.output_fmt)
        x = self.norm(x)
        return x


import os
import torch
import scipy
import numpy as np
from tqdm import tqdm
import pandas as pd
from skimage.segmentation import clear_border, find_boundaries
from skimage.io import imread
from scipy.ndimage.morphology import binary_dilation  as binary_dilation_scipy
from scipy.ndimage.measurements import center_of_mass                                                  
from skimage.morphology import disk, dilation, erosion, binary_erosion, binary_dilation
from skimage.measure import label, regionprops
from skimage.transform import resize
from skimage.feature import peak_local_max

from utils.util import load_data_from_dir, load_3d_images_from_dir, save_tif


#########################
# INSTANCE SEGMENTATION #
#########################
def create_instance_channels(cfg, data_type='train'):
    """Create training and validation new data with appropiate channels based on ``PROBLEM.INSTANCE_SEG.DATA_CHANNELS`` for instance
       segmentation.

       Parameters
       ----------
       cfg : YACS CN object
           Configuration.

	   data_type: str, optional
		   Wheter to create training or validation instance channels.

       Returns
       -------
       filenames: List of str
           Image paths.
    """

    assert data_type in ['train', 'val']

    f_name = load_data_from_dir if cfg.PROBLEM.NDIM == '2D' else load_3d_images_from_dir
    tag = "TRAIN" if data_type == "train" else "VAL"
    Y, _, _, filenames = f_name(getattr(cfg.DATA, tag).GT_PATH, check_drange=False, return_filenames=True)
    print("Creating Y_{} channels . . .".format(data_type))
    if isinstance(Y, list):
        for i in tqdm(range(len(Y))):
            Y[i] = labels_into_channels(Y[i], mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, save_dir=getattr(cfg.PATHS, tag+'_INSTANCE_CHANNELS_CHECK'),
                          fb_mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CONTOUR_MODE)
    else:
        Y = labels_into_channels(Y, mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, save_dir=getattr(cfg.PATHS, tag+'_INSTANCE_CHANNELS_CHECK'),
                   fb_mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CONTOUR_MODE)
    save_tif(Y, data_dir=getattr(cfg.DATA, tag).INSTANCE_CHANNELS_MASK_DIR, filenames=filenames, verbose=cfg.TEST.VERBOSE)
    X, _, _, filenames = f_name(getattr(cfg.DATA, tag).PATH, return_filenames=True)
    print("Creating X_{} channels . . .".format(data_type))
    save_tif(X, data_dir=getattr(cfg.DATA, tag).INSTANCE_CHANNELS_DIR, filenames=filenames, verbose=cfg.TEST.VERBOSE)
    
    # Save original X data with the labels 
    for i in range(min(3,len(X))):
        if isinstance(X, list):
            save_tif(X[i], getattr(cfg.PATHS, tag+'_INSTANCE_CHANNELS_CHECK'), filenames=['vol'+str(i)+".tif"], verbose=False)
        else:
            save_tif(np.expand_dims(X[i],0), getattr(cfg.PATHS, tag+'_INSTANCE_CHANNELS_CHECK'), filenames=['vol'+str(i)+".tif"], verbose=False)

def create_test_instance_channels(cfg):
    """Create test new data with appropiate channels based on ``PROBLEM.INSTANCE_SEG.DATA_CHANNELS`` for instance segmentation.

       Parameters
       ----------
       cfg : YACS CN object
           Configuration.
    """

    f_name = load_data_from_dir if cfg.PROBLEM.NDIM == '2D' else load_3d_images_from_dir

    if cfg.DATA.TEST.LOAD_GT:
        Y_test, _, _, test_filenames = f_name(cfg.DATA.TEST.GT_PATH, check_drange=False, return_filenames=True)
        print("Creating Y_test channels . . .")
        if isinstance(Y_test, list):
            for i in tqdm(range(len(Y_test))):
                Y_test[i] = labels_into_channels(Y_test[i], mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, save_dir=cfg.PATHS.TEST_INSTANCE_CHANNELS_CHECK,
                                            fb_mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CONTOUR_MODE)
        else:
            Y_test = labels_into_channels(Y_test, mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CHANNELS, save_dir=cfg.PATHS.TEST_INSTANCE_CHANNELS_CHECK,
                                     fb_mode=cfg.PROBLEM.INSTANCE_SEG.DATA_CONTOUR_MODE)
        save_tif(Y_test, data_dir=cfg.DATA.TEST.INSTANCE_CHANNELS_MASK_DIR, filenames=test_filenames, verbose=cfg.TEST.VERBOSE)

    print("Creating X_test channels . . .")
    X_test, _, _, test_filenames = f_name(cfg.DATA.TEST.PATH, return_filenames=True)
    save_tif(X_test, data_dir=cfg.DATA.TEST.INSTANCE_CHANNELS_DIR, filenames=test_filenames, verbose=cfg.TEST.VERBOSE)
    
    # Save original X data with the labels 
    for i in range(min(3,len(X_test))):
        if isinstance(X_test, list):
            save_tif(X_test[i], cfg.PATHS.TEST_INSTANCE_CHANNELS_CHECK, filenames=['vol'+str(i)+".tif"], verbose=True)
        else:
            save_tif(np.expand_dims(X_test[i],0), cfg.PATHS.TEST_INSTANCE_CHANNELS_CHECK, filenames=['vol'+str(i)+".tif"], verbose=True)

def labels_into_channels(data_mask, mode="BC", fb_mode="outer", save_dir=None):
    """Converts input semantic or instance segmentation data masks into different binary channels to train an instance segmentation
       problem. 

       Parameters
       ----------
       data_mask : 4D/5D Numpy array
           Data mask to create the new array from. It is expected to have just one channel. E.g. ``(10, 200, 1000, 1000, 1)``

       mode : str, optional
           Operation mode. Possible values: ``BC`` and ``BCD``.  ``BC`` corresponds to use binary segmentation+contour.
           ``BCD`` stands for binary segmentation+contour+distances.

       fb_mode : str, optional
          Mode of the find_boundaries function from ``scikit-image`` or "dense". More info in:
          `find_boundaries() <https://scikit-image.org/docs/stable/api/skimage.segmentation.html#skimage.segmentation.find_boundaries>`_. 
          Choose "dense" to label as contour every pixel that is not in ``B`` channel. 

       save_dir : str, optional
           Path to store samples of the created array just to debug it is correct.

       Returns
       -------
       new_mask : 5D Numpy array
           5D array with 3 channels instead of one. E.g. ``(10, 200, 1000, 1000, 3)``
    """

    assert data_mask.ndim in [5, 4]
    assert mode in ['BC', 'BCM', 'BCD', 'BD', 'BCDv2', 'Dv2', 'BDv2', 'BP']

    d_shape = 4 if data_mask.ndim == 5 else 3
    if mode in ['BCDv2', 'Dv2', 'BDv2']:
        c_number = 4
    elif mode in ['BCD', 'BCM']:
        c_number = 3
    elif mode in ['BC', 'BP', 'BD']:
        c_number = 2

    if 'D' in mode:
        dtype = np.float32  
    else:
        dtype = np.uint8

    new_mask = np.zeros(data_mask.shape[:d_shape] + (c_number,), dtype=dtype)
    for img in tqdm(range(data_mask.shape[0])):
        vol = data_mask[img,...,0].astype(np.int64)
        instances = np.unique(vol)
        instance_count = len(instances)

        # Background distance
        if 'Dv2' in mode:
            # Background distance
            vol_b_dist = np.invert(vol>0)
            vol_b_dist= scipy.ndimage.distance_transform_edt(vol_b_dist)
            vol_b_dist = np.max(vol_b_dist)-vol_b_dist
            new_mask[img,...,3] = vol_b_dist.copy()

        # Semantic mask
        if 'B' in mode and instance_count != 1:
            new_mask[img,...,0] = (vol>0).copy().astype(np.uint8)

        # Central points
        if 'P' in mode and instance_count != 1:
            coords = center_of_mass(vol>0, vol, instances[1:])
            coords = np.round(coords).astype(int)
            for coord in coords:
                if data_mask.ndim == 5:
                    z,y,x = coord
                    new_mask[img,z,y,x,1] = 1
                else:
                    y,x = coord
                    new_mask[img,y,x,1] = 1

            if data_mask.ndim == 5:
                for i in range(new_mask.shape[1]):                                                                                  
                    new_mask[img,i,...,1] = dilation(new_mask[img,i,...,1], disk(3)) 
            else:
                new_mask[img,...,1] = dilation(new_mask[img,...,1], disk(3))  

        # Contour
        if ('C' in mode or 'Dv2' in mode) and instance_count != 1: 
            f = "thick" if fb_mode == "dense" else fb_mode
            new_mask[img,...,1] = find_boundaries(vol, mode=f).astype(np.uint8)
            if fb_mode == "dense" and mode != "BCM":
                new_mask[img,...,1] = 1 - binary_dilation(new_mask[img,...,1], disk(1))
                new_mask[img,...,1] = 1 - ( (vol>0) * new_mask[img,...,1])
            if 'B' in mode:
                # Remove contours from segmentation maps
                new_mask[img,...,0][np.where(new_mask[img,...,1] == 1)] = 0
            if mode == "BCM":
                new_mask[img,...,2] = (vol>0).astype(np.uint8)

        if ('D' in mode or 'Dv2' in mode) and instance_count != 1:
            # Foreground distance
            new_mask[img,...,-1] = scipy.ndimage.distance_transform_edt(new_mask[img,...,0])
            props = regionprops(vol, new_mask[img,...,-1])
            max_values = np.zeros(vol.shape)
            for i in range(len(props)):
                max_values = np.where(vol==props[i].label, props[i].intensity_max, max_values)     
            new_mask[img,...,-1] = max_values - new_mask[img,...,-1]

    # Normalize and merge distance channels
    if 'Dv2' in mode:
        # Normalize background   
        b_min = np.min(new_mask[...,3])
        b_max = np.max(new_mask[...,3])    
        new_mask[...,3] = (new_mask[...,3]-b_min)/(b_max-b_min)

        if instance_count != 1: 
            # Normalize foreground  
            f_min = np.min(new_mask[...,2])
            f_max = np.max(new_mask[...,2])
            new_mask[...,2] = (new_mask[...,2]-f_min)/(f_max-f_min)

            new_mask[...,2] = new_mask[...,3] - new_mask[...,2]

            # The intersection of the channels is the contour channel, so set it to the maximum value 1
            new_mask[...,2][new_mask[...,1]>0] = 1

        if mode == "BCDv2":
            new_mask = new_mask[...,:3]
        elif mode == "BDv2":
            new_mask = new_mask[...,[0,-1]]   
        elif mode == "Dv2":
            new_mask = np.expand_dims(new_mask[...,2], -1)

    if save_dir is not None:
        os.makedirs(save_dir, exist_ok=True)
        suffix = []
        if mode == "Dv2":
            suffix.append('_distance.tif')
        else:
            suffix.append('_semantic.tif')
        if mode in ["BC", "BCM", "BCD", "BCDv2"]:
            suffix.append('_contour.tif')
            if mode in ["BCD", "BCDv2"]:
                suffix.append('_distance.tif')
            elif mode == "BCM":
                suffix.append('_binary_mask.tif')
        elif mode == "BP":
            suffix.append('_points.tif')
        elif mode in ["BDv2", "BD"]:
            suffix.append('_distance.tif')

        for i in range(min(3,len(new_mask))):
            for j in range(len(suffix)):
                aux = new_mask[i,...,j]
                aux = np.expand_dims(np.expand_dims(aux,-1),0)
                save_tif(aux, save_dir, filenames=['vol'+str(i)+suffix[j]], verbose=False)
                save_tif(np.expand_dims(data_mask[i],0), save_dir, filenames=['vol'+str(i)+"_y.tif"], verbose=False)
    return new_mask

#############
# DETECTION #
#############
def create_detection_masks(cfg, data_type='train'):
    """Create detection masks based on CSV files.

       Parameters
       ----------
       cfg : YACS CN object
           Configuration.

	   data_type: str, optional
		   Wheter to create train, validation or test masks.
    """

    assert data_type in ['train', 'val', 'test']

    if data_type == "train":
        tag = "TRAIN"
    elif data_type == "val":
        tag = "VAL"
    else:
        tag = "TEST"
    img_dir = getattr(cfg.DATA, tag).PATH
    label_dir = getattr(cfg.DATA, tag).GT_PATH
    out_dir = getattr(cfg.DATA, tag).DETECTION_MASK_DIR
    img_ids = sorted(next(os.walk(img_dir))[2])
    img_ext = '.'+img_ids[0].split('.')[-1]
    ids = sorted(next(os.walk(label_dir))[2])
    if len(img_ids) != len(ids):
        raise ValueError("Different number of CSV files and images found ({} vs {}). "
            "Please check that every image has one and only one CSV file".format(len(ids), len(img_ids)))
    if cfg.PROBLEM.NDIM == '2D':
        req_dim = 2 
        req_columns = ['axis-0', 'axis-1']
        req_columns_class = ['axis-0', 'axis-1', 'class']
    else:
        req_dim = 3
        req_columns = ['axis-0', 'axis-1', 'axis-2']
        req_columns_class = ['axis-0', 'axis-1', 'axis-2', 'class']

    classes = cfg.MODEL.N_CLASSES if cfg.MODEL.N_CLASSES > 2 else 1
    print("Creating {} detection masks . . .".format(data_type))
    for i in range(len(ids)):
        img_filename = os.path.splitext(ids[i])[0]+img_ext
        if not os.path.exists(os.path.join(out_dir, img_filename)):
            print("Attempting to create mask from CSV file: {}".format(os.path.join(label_dir, ids[i])))
            if not os.path.exists(os.path.join(img_dir, img_filename)):
                print("WARNING: The image seems to have different name than its CSV file. Using the CSV file that's "
                "in the same spot (within the CSV files list) where the image is in its own list of images. Check if it is correct!")
                img_filename = img_ids[i]
            print("Its respective image seems to be: {}".format(os.path.join(img_dir, img_filename)))
            
            df = pd.read_csv(os.path.join(label_dir, ids[i]))  
            img = imread(os.path.join(img_dir, img_filename))
            
            # Adjust shape
            img = np.squeeze(img)
            if cfg.PROBLEM.NDIM == '2D':
                if img.ndim == 2:
                    img = np.expand_dims(img, -1)
                else:
                    if img.shape[0] <= 3: img = img.transpose((1,2,0))   
            else: 
                if img.ndim == 3: 
                    img = np.expand_dims(img, -1)
                else:
                    if img.shape[0] <= 3: img = img.transpose((1,2,3,0))

            # Discard first index column to not have error if it is not sorted 
            p_number=df.iloc[: , 0].to_list()
            df = df.iloc[: , 1:]
            df = df.rename(columns=lambda x: x.strip()) # trim spaces in column names
            if len(df.columns) == req_dim+1:
                if not all(df.columns == req_columns_class):
                    raise ValueError("CSV columns need to be {}".format(req_columns_class))
            elif len(df.columns) == req_dim:
                if not all(df.columns == req_columns):
                    raise ValueError("CSV columns need to be {}".format(req_columns))
            else:
                raise ValueError("CSV file {} need to have {} or {} columns. Found {}"
                                .format(os.path.join(label_dir, ids[i]), req_dim, req_dim+1, len(df.columns)))

            # Convert them to int in case they are floats
            df['axis-0'] = df['axis-0'].astype('int')
            df['axis-1'] = df['axis-1'].astype('int')
            if cfg.PROBLEM.NDIM == '3D':
                df['axis-2'] = df['axis-2'].astype('int')
            
            df = df.sort_values(by=['axis-0']) 

            # Obtain the points 
            z_axis_point = df['axis-0']                                                                       
            y_axis_point = df['axis-1']    
            if cfg.PROBLEM.NDIM == '3D':                                                                   
                x_axis_point = df['axis-2']
            
            # Class column present
            if len(df.columns) == req_dim+1:
                df['class'] = df['class'].astype('int')
                class_point = np.array(df['class']) 

                uniq = np.sort(np.unique(class_point))        
                if uniq[0] != 1:
                    raise ValueError("Class number must start with 1")    
                if not all(uniq == np.array(range(1,classes+1))):
                    raise ValueError("Classes must be consecutive, e.g [1,2,3,4..]. Given {}".format(uniq))   
            else:
                if classes > 1:
                    raise ValueError("MODEL.N_CLASSES > 1 but no class specified in CSV files (4th column must have class info)")
                class_point = [1] * len(z_axis_point)

            # Create masks
            print("Creating all points . . .")
            mask = np.zeros((img.shape[:-1]+(classes,)), dtype=np.uint8)
            for j in tqdm(range(len(z_axis_point)), total=len(z_axis_point), leave=False):
                a0_coord = z_axis_point[j]
                a1_coord = y_axis_point[j]
                if cfg.PROBLEM.NDIM == '3D':
                    a2_coord = x_axis_point[j]
                c_point = class_point[j]-1

                if c_point+1 > mask.shape[-1]:
                    raise ValueError("Class {} detected while MODEL.N_CLASSES was set to {}. Please check it!"
                        .format(c_point+1, classes))

                # Paint the point
                if cfg.PROBLEM.NDIM == '3D':
                    if a0_coord < mask.shape[0] and a1_coord < mask.shape[1] and a2_coord < mask.shape[2]:
                        cpd = cfg.PROBLEM.DETECTION.CENTRAL_POINT_DILATION                                                                                                              
                        if 1 in mask[max(0,a0_coord-1):min(mask.shape[0],a0_coord+2),                                   
                                     max(0,a1_coord-1-cpd):min(mask.shape[1],a1_coord+2+cpd),                                   
                                     max(0,a2_coord-1-cpd):min(mask.shape[2],a2_coord+2+cpd), c_point]: 
                            print("WARNING: possible duplicated point in (3,9,9) neighborhood: coords {} , class {} "
                                  "(point number {} in CSV)".format((a0_coord,a1_coord,a2_coord), c_point, p_number[j]))                                                                                                                                            
                                                                                                                                
                        mask[a0_coord,a1_coord,a2_coord,c_point] = 1                                            
                        if a1_coord+1 < mask.shape[1]: mask[a0_coord,a1_coord+1,a2_coord,c_point] = 1       
                        if a1_coord-1 > 0: mask[a0_coord,a1_coord-1,a2_coord,c_point] = 1                   
                        if a2_coord+1 < mask.shape[2]: mask[a0_coord,a1_coord,a2_coord+1,c_point] = 1       
                        if a2_coord-1 > 0: mask[a0_coord,a1_coord,a2_coord-1,c_point] = 1     
                        if cfg.PROBLEM.DETECTION.CENTRAL_POINT_DILATION == 0:
                            if a1_coord+1 < mask.shape[1] and a2_coord+1 < mask.shape[2]: 
                                mask[a0_coord,a1_coord+1,a2_coord+1,c_point] = 1       
                            if a1_coord-1 > 0 and a2_coord-1 > 0: 
                                mask[a0_coord,a1_coord-1,a2_coord-1,c_point] = 1  
                            if a1_coord-1 > 0 and a2_coord+1 < mask.shape[2]: 
                                mask[a0_coord,a1_coord-1,a2_coord+1,c_point] = 1   
                            if a1_coord+1 < mask.shape[1] and a2_coord-1 > 0: 
                                mask[a0_coord,a1_coord+1,a2_coord-1,c_point] = 1             
                    else:  
                        print("WARNING: discarding point {} which seems to be out of shape: {}"
                              .format([a0_coord,a1_coord,a2_coord], img.shape))                                                                                                     
                else:
                    if a0_coord < mask.shape[0] and a1_coord < mask.shape[1]:                                                                                                              
                        if 1 in mask[max(0,a0_coord-4):min(mask.shape[0],a0_coord+5),                                   
                                    max(0,a1_coord-4):min(mask.shape[1],a1_coord+5), c_point]: 
                            print("WARNING: possible duplicated point in (9,9) neighborhood: coords {} , class {} "
                                  "(point number {} in CSV)".format((a0_coord,a1_coord), c_point, p_number[j]))                                                                                                                                            
                                                                                                                                
                        mask[a0_coord,a1_coord,c_point] = 1                                            
                        if a1_coord+1 < mask.shape[1]: mask[a0_coord,a1_coord+1,c_point] = 1       
                        if a1_coord-1 > 0: mask[a0_coord,a1_coord-1,c_point] = 1                                     
                    else:  
                        print("WARNING: discarding point {} which seems to be out of shape: {}"
                              .format([a0_coord,a1_coord], img.shape))     

            # Dilate the mask
            if cfg.PROBLEM.DETECTION.CENTRAL_POINT_DILATION > 0:
                if cfg.PROBLEM.NDIM == '2D': mask = np.expand_dims(mask,0)
                for k in range(mask.shape[0]): 
                    for ch in range(mask.shape[-1]):                                                                                  
                        mask[k,...,ch] = binary_dilation_scipy(mask[k,...,ch], iterations=1,  structure=disk(cfg.PROBLEM.DETECTION.CENTRAL_POINT_DILATION))                                                                                                                                                    
                if cfg.PROBLEM.NDIM == '2D': mask = mask[0]

            if cfg.PROBLEM.DETECTION.CHECK_POINTS_CREATED:
                print("Check points created to see if some of them are very close that create a large label") 
                error_found = False
                for ch in range(mask.shape[-1]):
                    _, index, counts = np.unique(label(clear_border(mask[...,ch])), return_counts=True, return_index=True)                     
                    # 0 is background so valid element is 1. We will compare that value with the rest                                                                         
                    ref_value = counts[1]                                                                                           
                    for k in range(2,len(counts)):                                                                                  
                        if abs(ref_value - counts[k]) > 5:                                                                          
                            point = np.unravel_index(index[k], mask[...,ch].shape)  
                            print("WARNING: There is a point (coords {}) with size very different from "
                                  "the rest. Maybe that cell has several labels: please check it! Normally all point "
                                  "have {} pixels but this one has {}.".format(point, ref_value, counts[k])) 
                            error_found = True

                if error_found:
                    raise ValueError("Duplicate points have been found so please check them before continuing. "
                                     "If you consider that the points are valid simply disable "
                                     "'PROBLEM.DETECTION.CHECK_POINTS_CREATED' so this check is not done again!")

            save_tif(np.expand_dims(mask,0), out_dir, [img_filename])
        else:
            print("Mask file {} found for CSV file: {}".format(os.path.join(out_dir, img_filename), 
                os.path.join(label_dir, ids[i])))

#######
# SSL #
#######
def create_ssl_source_data_masks(cfg, data_type='train'):
    """Create SSL source data.

       Parameters
       ----------
       cfg : YACS CN object
           Configuration.

	   data_type: str, optional
		   Wheter to create train, validation or test source data.
    """

    assert data_type in ['train', 'val', 'test']
    tag = data_type.upper()
    
    img_dir = getattr(cfg.DATA, tag).PATH
    out_dir = getattr(cfg.DATA, tag).SSL_SOURCE_DIR
    ids = sorted(next(os.walk(img_dir))[2])
    add_noise = True if cfg.PROBLEM.SELF_SUPERVISED.NOISE > 0 else False

    print("Creating {} SSL source. . .".format(data_type))
    for i in range(len(ids)):
        if not os.path.exists(os.path.join(out_dir, ids[i])):
            print("Crappifying file {} to create SSL source".format(os.path.join(img_dir, ids[i])))

            img = imread(os.path.join(img_dir, ids[i]))
            
            # Adjust shape
            img = np.squeeze(img)
            if cfg.PROBLEM.NDIM == '2D':
                if img.ndim == 2:
                    img = np.expand_dims(img, -1)
                else:
                    if img.shape[0] <= 3: img = img.transpose((1,2,0))   
            else: 
                if img.ndim == 3: 
                    img = np.expand_dims(img, -1)
                else:
                    min_val = min(img.shape)
                    channel_pos = img.shape.index(min_val)
                    if channel_pos != 3 and img.shape[channel_pos] <= 4:
                        new_pos = [x for x in range(4) if x != channel_pos]+[channel_pos,]
                        img = img.transpose(new_pos)

            img = crappify(img, resizing_factor=cfg.PROBLEM.SELF_SUPERVISED.RESIZING_FACTOR, 
                add_noise=add_noise, noise_level=cfg.PROBLEM.SELF_SUPERVISED.NOISE)

            save_tif(np.expand_dims(img,0), out_dir, [ids[i]])
        else:
            print("Source file {} found".format(os.path.join(img_dir, ids[i])))

def crappify(input_img, resizing_factor, add_noise=True, noise_level=None, Down_up=True):
    """
    Crappifies input image by adding Gaussian noise and downsampling and upsampling it so the resolution
    gets worsen. 

    input_img : 4D/5D Numpy array
        Data to be modified. E.g. ``(y, x, channels)`` if working with 2D images or
        ``(z, y, x, channels)`` if working with 3D.

    resizing_factor : floats
        Downsizing factor to reshape the image.

    add_noise : boolean, optional
        Indicating whether to add gaussian noise before applying the resizing.

    noise_level: float, optional
        Number between ``[0,1]`` indicating the std of the Gaussian noise N(0,std).

    Down_up : bool, optional
        Indicating whether to perform a final upsampling operation to obtain an image of the 
        same size as the original but with the corresponding loss of quality of downsizing and 
        upsizing.

    Returns
    -------
    img : 4D/5D Numpy array
        Train images. E.g. ``(y, x, channels)`` if working with 2D images or
        ``(z, y, x, channels)`` if working with 3D.
    """
    if input_img.ndim == 3:
        w, h, c = input_img.shape
        org_sz = (w, h)
    else:
        d, w, h, c = input_img.shape
        org_sz = (d, w, h)
        new_d = int(d / np.sqrt(resizing_factor))

    new_w = int(w / np.sqrt(resizing_factor))
    new_h = int(h / np.sqrt(resizing_factor))

    if input_img.ndim == 3:
        targ_sz = (new_w, new_h)
    else:
        targ_sz = (new_d, new_w, new_h)
    
    img = input_img.copy()
    if add_noise:
        img = add_gaussian_noise(img, noise_level)

    img = resize(img, targ_sz, order=1, mode='reflect',
                 clip=True, preserve_range=True, anti_aliasing=False)

    if Down_up:
        img = resize(img, org_sz, order=1, mode='reflect',
                        clip=True, preserve_range=True, anti_aliasing=False)

    return img.astype(input_img.dtype)

def add_gaussian_noise(image, percentage_of_noise):
    """
    Adds Gaussian noise to an input image. 

    Parameters
    ----------
    image : 3D Numpy array
        Image to be added Gaussian Noise with 0 mean and a certain std. E.g. ``(y, x, channels)``.

    percentage_of_noise : float
        percentage of the maximum value of the image that will be used as the std of the Gaussian Noise 
        distribution.

    Returns
    -------
    out : 3D Numpy array
        Transformed image. E.g. ``(y, x, channels)``.
    """
    max_value=np.max(image)
    noise_level=percentage_of_noise*max_value
    noise = np.random.normal(loc=0, scale=noise_level, size=image.shape)
    noisy_img=np.clip(image+noise, 0, max_value).astype(image.dtype)
    return noisy_img

################
# SEMANTIC SEG #
################
def calculate_2D_volume_prob_map(Y, Y_path=None, w_foreground=0.94, w_background=0.06, save_dir=None):
    """Calculate the probability map of the given 2D data.

       Parameters
       ----------
       Y : 4D Numpy array
           Data to calculate the probability map from. E. g. ``(num_of_images, y, x, channel)``

       Y_path : str, optional
           Path to load the data from in case ``Y=None``.

       w_foreground : float, optional
           Weight of the foreground. This value plus ``w_background`` must be equal ``1``.

       w_background : float, optional
           Weight of the background. This value plus ``w_foreground`` must be equal ``1``.

       save_dir : str, optional
           Path to the file where the probability map will be stored.

       Raises
       ------
       ValueError
           if ``Y`` does not have 4 dimensions.

       ValueError
           if ``w_foreground + w_background > 1``.

       Returns
       -------
       Array : Str or 4D Numpy array
           Path where the probability map/s is/are stored if ``Y_path`` was given and there are images of different
           shapes. Otherwise, an array that represents the probability map of ``Y`` or all loaded data files from
           ``Y_path`` will be returned.
    """

    if Y is not None:
        if Y.ndim != 4:
            raise ValueError("'Y' must be a 4D Numpy array")

    if Y is None and Y_path is None:
        raise ValueError("'Y' or 'Y_path' need to be provided")

    if Y is not None:
        prob_map = np.copy(Y).astype(np.float32)
        l = prob_map.shape[0]
        channels = prob_map.shape[-1]
        v = np.max(prob_map)
    else:
        prob_map, _, _ = load_data_from_dir(Y_path)
        l = len(prob_map)
        channels = prob_map[0].shape[-1]
        v = np.max(prob_map[0])

    if isinstance(prob_map, list):
        first_shape = prob_map[0][0].shape
    else:
        first_shape = prob_map[0].shape

    print("Connstructing the probability map . . .")
    maps = []
    diff_shape = False
    for i in tqdm(range(l)):
        if isinstance(prob_map, list):
            _map = prob_map[i][0].copy().astype(np.float32)
        else:
            _map = prob_map[i].copy().astype(np.float32)

        for k in range(channels):
            # Remove artifacts connected to image border
            _map[:,:,k] = clear_border(_map[:,:,k])

            foreground_pixels = (_map[:,:,k] == v).sum()
            background_pixels = (_map[:,:,k] == 0).sum()

            if foreground_pixels == 0:
                _map[:,:,k][np.where(_map[:,:,k] == v)] = 0
            else:
                _map[:,:,k][np.where(_map[:,:,k] == v)] = w_foreground/foreground_pixels
            if background_pixels == 0:
                _map[:,:,k][np.where(_map[:,:,k] == 0)] = 0
            else:
                _map[:,:,k][np.where(_map[:,:,k] == 0)] = w_background/background_pixels

            # Necessary to get all probs sum 1
            s = _map[:,:,k].sum()
            if s == 0:
                t = 1
                for x in _map[:,:,k].shape: t *=x
                _map[:,:,k].fill(1/t)
            else:
                _map[:,:,k] = _map[:,:,k]/_map[:,:,k].sum()

        if first_shape != _map.shape: diff_shape = True
        maps.append(_map)

    if save_dir is not None:
        os.makedirs(save_dir, exist_ok=True)

        if not diff_shape:
            for i in range(len(maps)):
                maps[i] = np.expand_dims(maps[i], 0)
            maps = np.concatenate(maps)
            print("Saving the probability map in {}".format(save_dir))
            np.save(os.path.join(save_dir, 'prob_map.npy'), maps)
            return maps
        else:
            print("As the files loaded have different shapes, the probability map for each one will be stored"
                  " separately in {}".format(save_dir))
            d = len(str(l))
            for i in range(l):
                f = os.path.join(save_dir, 'prob_map'+str(i).zfill(d)+'.npy')
                np.save(f, maps[i])
            return save_dir

def calculate_3D_volume_prob_map(Y, Y_path=None, w_foreground=0.94, w_background=0.06, save_dir=None):
    """Calculate the probability map of the given 3D data.

       Parameters
       ----------
       Y : 5D Numpy array
           Data to calculate the probability map from. E. g. ``(num_subvolumes, z, y, x, channel)``

       Y_path : str, optional
           Path to load the data from in case ``Y=None``.

       w_foreground : float, optional
           Weight of the foreground. This value plus ``w_background`` must be equal ``1``.

       w_background : float, optional
           Weight of the background. This value plus ``w_foreground`` must be equal ``1``.

       save_dir : str, optional
           Path to the directory where the probability map will be stored.

       Returns
       -------
       Array : Str or 5D Numpy array
           Path where the probability map/s is/are stored if ``Y_path`` was given and there are images of different
           shapes. Otherwise, an array that represents the probability map of ``Y`` or all loaded data files from
           ``Y_path`` will be returned.

       Raises
       ------
       ValueError
           if ``Y`` does not have 5 dimensions.
       ValueError
           if ``w_foreground + w_background > 1``.
    """

    if Y is not None:
        if Y.ndim != 5:
            raise ValueError("'Y' must be a 5D Numpy array")

    if Y is None and Y_path is None:
        raise ValueError("'Y' or 'Y_path' need to be provided")

    if Y is not None:
        prob_map = np.copy(Y).astype(np.float32)
        l = prob_map.shape[0]
        channels = prob_map.shape[-1]
        v = np.max(prob_map)
    else:
        prob_map, _, _ = load_3d_images_from_dir(Y_path)
        l = len(prob_map)
        channels = prob_map[0].shape[-1]
        v = np.max(prob_map[0])

    if isinstance(prob_map, list):
        first_shape = prob_map[0][0].shape
    else:
        first_shape = prob_map[0].shape

    print("Constructing the probability map . . .")
    maps = []
    diff_shape = False
    for i in range(l):
        if isinstance(prob_map, list):
            _map = prob_map[i][0].copy().astype(np.float64)
        else:
            _map = prob_map[i].copy().astype(np.float64)

        for k in range(channels):
            for j in range(_map.shape[0]):
                # Remove artifacts connected to image border
                _map[j,:,:,k] = clear_border(_map[j,:,:,k])
            foreground_pixels = (_map[:,:,:,k] == v).sum()
            background_pixels = (_map[:,:,:,k] == 0).sum()

            if foreground_pixels == 0:
                _map[:,:,:,k][np.where(_map[:,:,:,k] == v)] = 0
            else:
                _map[:,:,:,k][np.where(_map[:,:,:,k] == v)] = w_foreground/foreground_pixels
            if background_pixels == 0:
                _map[:,:,:,k][np.where(_map[:,:,:,k] == 0)] = 0
            else:
                _map[:,:,:,k][np.where(_map[:,:,:,k] == 0)] = w_background/background_pixels

            # Necessary to get all probs sum 1
            s = _map[:,:,:,k].sum()
            if s == 0:
                t = 1
                for x in _map[:,:,:,k].shape: t *=x
                _map[:,:,:,k].fill(1/t)
            else:
                _map[:,:,:,k] = _map[:,:,:,k]/_map[:,:,:,k].sum()

        if first_shape != _map.shape: diff_shape = True
        maps.append(_map)

    if save_dir is not None:
        os.makedirs(save_dir, exist_ok=True)
        if not diff_shape:
            for i in range(len(maps)):
                maps[i] = np.expand_dims(maps[i], 0)
            maps = np.concatenate(maps)
            print("Saving the probability map in {}".format(save_dir))
            np.save(os.path.join(save_dir, 'prob_map.npy'), maps)
            return maps
        else:
            print("As the files loaded have different shapes, the probability map for each one will be stored "
                  "separately in {}".format(save_dir))
            d = len(str(l))
            for i in range(l):
                f = os.path.join(save_dir, 'prob_map'+str(i).zfill(d)+'.npy')
                np.save(f, maps[i])
            return save_dir

###########
# GENERAL #
###########
def norm_range01(x, dtype=np.float32):
    norm_steps = {}
    norm_steps['orig_dtype'] = x.dtype
    if x.dtype == np.uint8:
        x = x/255
        norm_steps['div_255'] = 1
    else:
        if np.max(x) > 255:
            norm_steps['reduced_{}'.format(x.dtype)] = 1
            x = reduce_dtype(x, 0, 65535, out_min=0, out_max=1, out_type=dtype)
        elif np.max(x) > 2:
            x = x/255
            norm_steps['div_255'] = 1

    x = x.astype(dtype)
    return x, norm_steps

def undo_norm_range01(x, xnorm):
    if 'div_255' in xnorm:
        x = (x*255).astype(np.uint8)
    reductions = [key for key, value in xnorm.items() if 'reduced' in key.lower()]
    if len(reductions)>0:
        reductions = reductions[0]
        reductions = reductions.replace('reduced_','')
        x = (x*65535).astype(eval("np.{}".format(reductions) ))
    return x

def reduce_dtype(x, x_min, x_max, out_min=0, out_max=1, out_type=np.float32):
    return ((np.array((x-x_min)/(x_max-x_min))*(out_max-out_min))+out_min).astype(out_type)

def normalize(data, means, stds, out_type="float32"):
    numpy_torch_dtype_dict = {
        "bool"       : [torch.bool, bool],
        "uint8"      : [torch.uint8, np.uint8],
        "int8"       : [torch.int8, np.int8],
        "int16"      : [torch.int16, np.int16],
        "int32"      : [torch.int32, np.int32],
        "int64"      : [torch.int64, np.int64],
        "float16"    : [torch.float16, np.float16],
        "float32"    : [torch.float32, np.float32],
        "float64"    : [torch.float64, np.float64],
        "complex64"  : [torch.complex64, np.complex64],
        "complex128" : [torch.complex128, np.complex128],
    }
    if torch.is_tensor(data):
        return ((data - means) / stds).to(numpy_torch_dtype_dict[out_type][0])
    else:
        return ((data - means) / stds).astype(numpy_torch_dtype_dict[out_type][1])

def denormalize(data, means, stds, out_type="float32"):
    numpy_torch_dtype_dict = {
        "bool"       : [torch.bool, bool],
        "uint8"      : [torch.uint8, np.uint8],
        "int8"       : [torch.int8, np.int8],
        "int16"      : [torch.int16, np.int16],
        "int32"      : [torch.int32, np.int32],
        "int64"      : [torch.int64, np.int64],
        "float16"    : [torch.float16, np.float16],
        "float32"    : [torch.float32, np.float32],
        "float64"    : [torch.float64, np.float64],
        "complex64"  : [torch.complex64, np.complex64],
        "complex128" : [torch.complex128, np.complex128],
    }
    if torch.is_tensor(data):
        return ((data * stds) + means).to(numpy_torch_dtype_dict[out_type][0])
    else:
        return ((data * stds) + means).astype(numpy_torch_dtype_dict[out_type][1])


import os
import sys
import numpy as np
from distutils.sysconfig import get_python_inc
from setuptools import setup, Extension, find_namespace_packages

requirements = [
     'imgaug>=0.4.0',
     'matplotlib>=3.7.1',
     'scikit-learn>=1.2.2',
     'pydot>=1.4.2',
     'yacs>=0.1.8',
     'tqdm>=4.65.0',
     'scikit-image>=0.21.0',
     'edt>=2.3.1',
     'fill-voids>=2.0.5',
     'opencv-python>=4.7.0.72',
     'pandas>=2.0.2',
     'torchinfo>=1.8.0',
     'torchmetrics>=1.0.3',
     'tensorboardX>=2.6.2.2',
     'timm>=0.9.5',
     'h5py>=3.10.0',
]


def getInclude():
    dirName = get_python_inc()
    return [dirName, os.path.dirname(dirName), np.get_include()]


def setup_package():
    __version__ = '3.0'
    url = 'https://github.com/danifranco/BiaPy'

    setup(name='biapy',
          description='Bioimage analysis pipelines in Python',
          version=__version__,
          url=url,
          license='MIT',
          author='BiaPy Contributors',
          install_requires=requirements,
          include_dirs=getInclude(),
          packages=find_namespace_packages(),
          include_package_data=True,
          )


if __name__ == '__main__':
    # pip install --editable .
    setup_package()



"""Copied from https://github.com/stardist/stardist"""

import numpy as np

from tqdm import tqdm
from scipy.optimize import linear_sum_assignment
from collections import namedtuple
import pandas as pd
import networkx as nx

matching_criteria = dict()

# Copied from https://github.com/CSBDeep/CSBDeep/blob/master/csbdeep/utils/utils.py
def _raise(e):
    if isinstance(e, BaseException):
        raise e
    else:
        raise ValueError(e)

def label_are_sequential(y):
    """ returns true if y has only sequential labels from 1... """
    labels = np.unique(y)
    return (set(labels)-{0}) == set(range(1,1+labels.max()))


def is_array_of_integers(y):
    return isinstance(y,np.ndarray) and np.issubdtype(y.dtype, np.integer)


def _check_label_array(y, name=None, check_sequential=False):
    err = ValueError("{label} must be an array of {integers}.".format(
        label = 'labels' if name is None else name,
        integers = ('sequential ' if check_sequential else '') + 'non-negative integers',
    ))
    is_array_of_integers(y) or _raise(err)
    if len(y) == 0:
        return True
    if check_sequential:
        label_are_sequential(y) or _raise(err)
    else:
        y.min() >= 0 or _raise(err)
    return True


def label_overlap(x, y, check=True):
    if check:
        _check_label_array(x,'x',True)
        _check_label_array(y,'y',True)
        x.shape == y.shape or _raise(ValueError("x and y must have the same shape"))
    return _label_overlap(x, y)

def _label_overlap(x, y):
    x = x.ravel()
    y = y.ravel()
    overlap = np.zeros((1+x.max(),1+y.max()), dtype=np.uint)
    for i in range(len(x)):
        overlap[x[i],y[i]] += 1
    return overlap

def _safe_divide(x,y, eps=1e-10):
    """computes a safe divide which returns 0 if y is zero"""
    if np.isscalar(x) and np.isscalar(y):
        return x/y if np.abs(y)>eps else 0.0
    else:
        out = np.zeros(np.broadcast(x,y).shape, np.float32)
        np.divide(x,y, out=out, where=np.abs(y)>eps)
        return out


def intersection_over_union(overlap):
    _check_label_array(overlap,'overlap')
    if np.sum(overlap) == 0:
        return overlap
    n_pixels_pred = np.sum(overlap, axis=0, keepdims=True)
    n_pixels_true = np.sum(overlap, axis=1, keepdims=True)
    return _safe_divide(overlap, (n_pixels_pred + n_pixels_true - overlap))

matching_criteria['iou'] = intersection_over_union


def intersection_over_true(overlap):
    _check_label_array(overlap,'overlap')
    if np.sum(overlap) == 0:
        return overlap
    n_pixels_true = np.sum(overlap, axis=1, keepdims=True)
    return _safe_divide(overlap, n_pixels_true)

matching_criteria['iot'] = intersection_over_true


def intersection_over_pred(overlap):
    _check_label_array(overlap,'overlap')
    if np.sum(overlap) == 0:
        return overlap
    n_pixels_pred = np.sum(overlap, axis=0, keepdims=True)
    return _safe_divide(overlap, n_pixels_pred)

matching_criteria['iop'] = intersection_over_pred


def precision(tp,fp,fn):
    return tp/(tp+fp) if tp > 0 else 0
def recall(tp,fp,fn):
    return tp/(tp+fn) if tp > 0 else 0
def accuracy(tp,fp,fn):
    # also known as "average precision" (?)
    # -> https://www.kaggle.com/c/data-science-bowl-2018#evaluation
    return tp/(tp+fp+fn) if tp > 0 else 0
def f1(tp,fp,fn):
    # also known as "dice coefficient"
    return (2*tp)/(2*tp+fp+fn) if tp > 0 else 0


def matching(y_true, y_pred, thresh=0.5, criterion='iou', report_matches=False):
    """Calculate detection/instance segmentation metrics between ground truth and predicted label images.

    Currently, the following metrics are implemented:

    'fp', 'tp', 'fn', 'precision', 'recall', 'accuracy', 'f1', 'criterion', 'thresh', 'n_true', 'n_pred', 'mean_true_score', 'mean_matched_score', 'panoptic_quality'

    Corresponding objects of y_true and y_pred are counted as true positives (tp), false positives (fp), and false negatives (fn)
    whether their intersection over union (IoU) >= thresh (for criterion='iou', which can be changed)

    * mean_matched_score is the mean IoUs of matched true positives

    * mean_true_score is the mean IoUs of matched true positives but normalized by the total number of GT objects

    * panoptic_quality defined as in Eq. 1 of Kirillov et al. "Panoptic Segmentation", CVPR 2019

    Parameters
    ----------
    y_true: ndarray
        ground truth label image (integer valued)
    y_pred: ndarray
        predicted label image (integer valued)
    thresh: float
        threshold for matching criterion (default 0.5)
    criterion: string
        matching criterion (default IoU)
    report_matches: bool
        if True, additionally calculate matched_pairs and matched_scores (note, that this returns even gt-pred pairs whose scores are below  'thresh')

    Returns
    -------
    Matching object with different metrics as attributes

    Examples
    --------
    >>> y_true = np.zeros((100,100), np.uint16)
    >>> y_true[10:20,10:20] = 1
    >>> y_pred = np.roll(y_true,5,axis = 0)

    >>> stats = matching(y_true, y_pred)
    >>> print(stats)
    Matching(criterion='iou', thresh=0.5, fp=1, tp=0, fn=1, precision=0, recall=0, accuracy=0, f1=0, n_true=1, n_pred=1, mean_true_score=0.0, mean_matched_score=0.0, panoptic_quality=0.0)

    """
    _check_label_array(y_true,'y_true')
    _check_label_array(y_pred,'y_pred')
    y_true.shape == y_pred.shape or _raise(ValueError("y_true ({y_true.shape}) and y_pred ({y_pred.shape}) have different shapes".format(y_true=y_true, y_pred=y_pred)))
    criterion in matching_criteria or _raise(ValueError("Matching criterion '%s' not supported." % criterion))
    if thresh is None: thresh = 0
    thresh = float(thresh) if np.isscalar(thresh) else map(float,thresh)

    y_true, _, map_rev_true = relabel_sequential(y_true)
    y_pred, _, map_rev_pred = relabel_sequential(y_pred)

    overlap = label_overlap(y_true, y_pred, check=False)
    scores = matching_criteria[criterion](overlap)
    assert 0 <= np.min(scores) <= np.max(scores) <= 1

    # ignoring background
    scores = scores[1:,1:]
    n_true, n_pred = scores.shape
    n_matched = min(n_true, n_pred)

    def _single(thr):
        not_trivial = n_matched > 0 and np.any(scores >= thr)
        if not_trivial:
            # compute optimal matching with scores as tie-breaker
            costs = -(scores >= thr).astype(float) - scores / (2*n_matched)
            true_ind, pred_ind = linear_sum_assignment(costs)
            assert n_matched == len(true_ind) == len(pred_ind)
            match_ok = scores[true_ind,pred_ind] >= thr
            tp = np.count_nonzero(match_ok)
        else:
            tp = 0
        fp = n_pred - tp
        fn = n_true - tp
        # assert tp+fp == n_pred
        # assert tp+fn == n_true

        # the score sum over all matched objects (tp)
        sum_matched_score = np.sum(scores[true_ind,pred_ind][match_ok]) if not_trivial else 0.0

        # the score average over all matched objects (tp)
        mean_matched_score = _safe_divide(sum_matched_score, tp)
        # the score average over all gt/true objects
        mean_true_score    = _safe_divide(sum_matched_score, n_true)
        panoptic_quality   = _safe_divide(sum_matched_score, tp+fp/2+fn/2)

        stats_dict = dict (
            criterion          = criterion,
            thresh             = thr,
            fp                 = fp,
            tp                 = tp,
            fn                 = fn,
            precision          = precision(tp,fp,fn),
            recall             = recall(tp,fp,fn),
            accuracy           = accuracy(tp,fp,fn),
            f1                 = f1(tp,fp,fn),
            n_true             = n_true,
            n_pred             = n_pred,
            mean_true_score    = mean_true_score,
            mean_matched_score = mean_matched_score,
            panoptic_quality   = panoptic_quality,
        )
        if bool(report_matches):
            if not_trivial:
                stats_dict.update (
                    # int() to be json serializable
                    matched_pairs  = tuple((int(map_rev_true[i]),int(map_rev_pred[j])) for i,j in zip(1+true_ind,1+pred_ind)),
                    matched_scores = tuple(scores[true_ind,pred_ind]),
                    matched_tps    = tuple(map(int,np.flatnonzero(match_ok))),
                    pred_ids       = tuple(map_rev_pred),        
                    gt_ids         = tuple(map_rev_true),
                )
            else:
                stats_dict.update (
                    matched_pairs  = (),
                    matched_scores = (),
                    matched_tps    = (),
                    pred_ids       = (), 
                    gt_ids         = (),
                )
        return stats_dict

    return _single(thresh) if np.isscalar(thresh) else tuple(map(_single,thresh))


def wrapper_matching_dataset_lazy(stats_all, thresh, criterion='iou', by_image=False):

    expected_keys = set(('fp', 'tp', 'fn', 'precision', 'recall', 'accuracy', 'f1', 'criterion', 'thresh', 'n_true', 'n_pred', 'mean_true_score', 'mean_matched_score', 'panoptic_quality'))

    # accumulate results over all images for each threshold separately
    n_images, n_threshs = len(stats_all), len(thresh)
    single_thresh = True if n_threshs == 1 else False
    accumulate = [{} for _ in range(n_threshs)]
    for stats in stats_all:
        for i, s in enumerate(stats):
            acc = accumulate[i]
            for item in s.items():
                k, v = item
                if k == 'mean_true_score' and not bool(by_image):
                    # convert mean_true_score to "sum_matched_score"
                    acc[k] = acc.setdefault(k,0) + v * s['n_true']
                else:
                    try:
                        acc[k] = acc.setdefault(k,0) + v
                    except TypeError:
                        pass

    # normalize/compute 'precision', 'recall', 'accuracy', 'f1'
    for thr,acc in zip(thresh,accumulate):
        acc['criterion'] = criterion
        acc['thresh'] = thr
        acc['by_image'] = bool(by_image)
        if bool(by_image):
            for k in ('precision', 'recall', 'accuracy', 'f1', 'mean_true_score', 'mean_matched_score', 'panoptic_quality'):
                acc[k] /= n_images
        else:
            tp, fp, fn, n_true = acc['tp'], acc['fp'], acc['fn'], acc['n_true']
            sum_matched_score = acc['mean_true_score']

            mean_matched_score = _safe_divide(sum_matched_score, tp)
            mean_true_score    = _safe_divide(sum_matched_score, n_true)
            panoptic_quality   = _safe_divide(sum_matched_score, tp+fp/2+fn/2)

            acc.update(
                precision          = precision(tp,fp,fn),
                recall             = recall(tp,fp,fn),
                accuracy           = accuracy(tp,fp,fn),
                f1                 = f1(tp,fp,fn),
                mean_true_score    = mean_true_score,
                mean_matched_score = mean_matched_score,
                panoptic_quality   = panoptic_quality,
            )

    accumulate = tuple(namedtuple('DatasetMatching',acc.keys())(*acc.values()) for acc in accumulate)
    return accumulate[0] if single_thresh else accumulate

def wrapper_matching_segCompare(stats_all):
    expected_keys = ['number_of_cells', 'correct_segmentations', 'oversegmentation_rate', 'undersegmentation_rate', 'missing_rate']

    accumulated_values = dict.fromkeys(expected_keys, 0)

    for key in expected_keys:
        for stat in stats_all:
            accumulated_values[key] = accumulated_values[key] + stat[key]
        accumulated_values[key] = accumulated_values[key]/len(stats_all)
    return accumulated_values

# copied from scikit-image master for now (remove when part of a release)
def relabel_sequential(label_field, offset=1):
    """Relabel arbitrary labels to {`offset`, ... `offset` + number_of_labels}.

    This function also returns the forward map (mapping the original labels to
    the reduced labels) and the inverse map (mapping the reduced labels back
    to the original ones).

    Parameters
    ----------
    label_field : numpy array of int, arbitrary shape
        An array of labels, which must be non-negative integers.
    offset : int, optional
        The return labels will start at `offset`, which should be
        strictly positive.

    Returns
    -------
    relabeled : numpy array of int, same shape as `label_field`
        The input label field with labels mapped to
        {offset, ..., number_of_labels + offset - 1}.
        The data type will be the same as `label_field`, except when
        offset + number_of_labels causes overflow of the current data type.
    forward_map : numpy array of int, shape ``(label_field.max() + 1,)``
        The map from the original label space to the returned label
        space. Can be used to re-apply the same mapping. See examples
        for usage. The data type will be the same as `relabeled`.
    inverse_map : 1D numpy array of int, of length offset + number of labels
        The map from the new label space to the original space. This
        can be used to reconstruct the original label field from the
        relabeled one. The data type will be the same as `relabeled`.

    Notes
    -----
    The label 0 is assumed to denote the background and is never remapped.

    The forward map can be extremely big for some inputs, since its
    length is given by the maximum of the label field. However, in most
    situations, ``label_field.max()`` is much smaller than
    ``label_field.size``, and in these cases the forward map is
    guaranteed to be smaller than either the input or output images.

    Examples
    --------
    >>> from skimage.segmentation import relabel_sequential
    >>> label_field = np.array([1, 1, 5, 5, 8, 99, 42])
    >>> relab, fw, inv = relabel_sequential(label_field)
    >>> relab
    array([1, 1, 2, 2, 3, 5, 4])
    >>> fw
    array([0, 1, 0, 0, 0, 2, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5])
    >>> inv
    array([ 0,  1,  5,  8, 42, 99])
    >>> (fw[label_field] == relab).all()
    True
    >>> (inv[relab] == label_field).all()
    True
    >>> relab, fw, inv = relabel_sequential(label_field, offset=5)
    >>> relab
    array([5, 5, 6, 6, 7, 9, 8])
    """
    offset = int(offset)
    if offset <= 0:
        raise ValueError("Offset must be strictly positive.")
    if np.min(label_field) < 0:
        raise ValueError("Cannot relabel array that contains negative values.")
    max_label = int(label_field.max()) # Ensure max_label is an integer
    if not np.issubdtype(label_field.dtype, np.integer):
        new_type = np.min_scalar_type(max_label)
        label_field = label_field.astype(new_type)
    labels = np.unique(label_field)
    labels0 = labels[labels != 0]
    new_max_label = offset - 1 + len(labels0)
    new_labels0 = np.arange(offset, new_max_label + 1)
    output_type = label_field.dtype
    required_type = np.min_scalar_type(new_max_label)
    if np.dtype(required_type).itemsize > np.dtype(label_field.dtype).itemsize:
        output_type = required_type
    forward_map = np.zeros(max_label + 1, dtype=output_type)
    forward_map[labels0] = new_labels0
    inverse_map = np.zeros(new_max_label + 1, dtype=output_type)
    inverse_map[offset:] = labels0
    relabeled = forward_map[label_field]
    return relabeled, forward_map, inverse_map
    
    
def segmentation_state(row):
    """Function from https://mosaic.gitlabpages.inria.fr/publications/seg_compare/
    used in match_using_segCompare function                                                    

       Inputs
       ----------
       row : pandas dataframe row
           
       Returns
       -------
       string with the segmentation state of a cell based on the row

    """

    if len(row.reference) == 0:
        return 'background'
    elif len(row.target) == 0:
        return 'missing'
    elif len(row.target) == 1:
        if len(row.reference) == 1:
            return 'one-to-one'
        else:
            return 'under-segmentation'
    else:
        if len(row.reference) == 1:
            return 'over-segmentation'
        else:
            return 'misc.'
    
def match_using_segCompare(y_true, y_pred, BACKGROUND_LABEL=0, thresh_background=0.5):

    """Calcualte over-segmentation and under-segmentation rates based on the paper
    under-segmentation rates based on the paper entitled "Assessment of deep learning algorithms
    for 3D instance segmentation of confocal image datasets" by A. Kar et al 
    (https://doi.org/10.1101/2021.06.09.447748)
     
    Their code is available at https://mosaic.gitlabpages.inria.fr/publications/seg_compare/
    but uses some libraries that are not compatible with the python version used in this 
    repository.                                                      

       Inputs
       ----------
       y_true : 4D Numpy array
           Ground Truth data. E.g. ``(img_number, x, y, channels)``.
           
       y_pred : 4D Numpy array
           Predicted data  E.g. ``(img_number, x, y, channels)``.

      Returns
      -------
      stats_dict: Dictionary with the following info:
          % correct segmentations: float
              one-to-one associations
          % undersegmentation: float
              one-to-many associations
          % oversegmentation: float
              many-to-one associations
          % misc: float
              many-to-many associations
          % missing: float
              if ground truth cell is associated is not in predicted image
          % background: float
              if predicted cell is associated with ground truth background

    """
    
    #Number of labels
    groundTruthLabelsNum = np.size(np.unique(y_true))
    predictedCellsNum = np.size(np.unique(y_pred))

    #Initialize variables
    groundTruthCellVolume = np.zeros([groundTruthLabelsNum])
    predictedCellVolume = np.zeros([predictedCellsNum])

    #Relabel (we assume that background is the lowest value in image)
    if min(np.unique(y_true)) != 0:
        y_true = relabel_sequential(y_true, 1)[0]-1
    else:
        y_true = relabel_sequential(y_true, 1)[0]
    if min(np.unique(y_pred)) != 0:
        y_pred = relabel_sequential(y_pred, 1)[0]-1
    else:
        y_pred = relabel_sequential(y_pred, 1)[0]  

    predictedCells = np.unique(y_pred)
    groundTruthCells = np.unique(y_true)

    #Calculate unions, intersections, jaccardIndex, volumes and assymetric inclusion indices
    #VJI is not strictly neccessary, could be removed or maybe hidden w/ an if statement.

    for predictedCell in predictedCells:
        #Calculate GroundTruth cell and predictedCellVolume Volume
        predictedCellVolume[predictedCell] = np.count_nonzero(y_pred==predictedCell)

    df_target = pd.DataFrame(columns = ['target', 'reference', 'target_in_reference'])
    df_reference = pd.DataFrame(columns = ['target', 'reference', 'reference_in_target'])

    for groundTruthCell in groundTruthCells:
        gtCell = y_true==groundTruthCell

        validSlices = [np.count_nonzero(gtCell[slice, :, :])>0 for slice in range(np.shape(gtCell)[0])]
        
        gtCell = gtCell[validSlices, :, :]
        groundTruthCellVolume[groundTruthCell] = np.count_nonzero(y_true[validSlices, :, :]==groundTruthCell)
        matchingCells = np.unique(y_pred[(y_true==groundTruthCell)])

        for predictedCell in matchingCells:
        
            pCell =  y_pred[validSlices, :, :]==predictedCell
        
            #Calculate union and intersection
            intersection = (gtCell & pCell).sum()

            df_target.loc[len(df_target.index)] = [predictedCell, groundTruthCell, intersection/predictedCellVolume[predictedCell]]
            df_reference.loc[len(df_reference.index)] = [predictedCell, groundTruthCell, intersection/groundTruthCellVolume[groundTruthCell]]


    ### - Solve the background associations - ###
    # target --> reference background
    target_background = df_target.target.loc[(df_target.reference == BACKGROUND_LABEL) &
                                             (df_target.target_in_reference >= thresh_background)].to_list()

    # reference --> target background
    reference_background = df_reference.reference.loc[(df_reference.target == BACKGROUND_LABEL) &
                                                      (df_reference.reference_in_target >= thresh_background)].to_list()

    # - Add the background (make sure we remove them for the clique associations)
    target_background.append(BACKGROUND_LABEL)
    reference_background.append(BACKGROUND_LABEL)

    # - Remove the cells associated with the background + backgrounds
    df_target = df_target.loc[~((df_target.target.isin(target_background)) |
                            (df_target.reference == BACKGROUND_LABEL))].copy()
    df_reference = df_reference.loc[~((df_reference.reference.isin(reference_background)) |
                                  (df_reference.target == BACKGROUND_LABEL))].copy()
                       
    ### - Get the 1 <--> 1 associations - ###
    # - Associate each reference/target cell with the target/reference cell in which it is the most included
    df_target = df_target.loc[df_target.groupby('target')['target_in_reference'].idxmax()]
    df_reference = df_reference.loc[df_reference.groupby('reference')['reference_in_target'].idxmax()]

    # - Convert in dict
    target_in_reference = df_target[['target', 'reference']].set_index('target').to_dict()['reference']
    reference_in_target = df_reference[['reference', 'target']].set_index('reference').to_dict()['target']

    ### - Build the associations (bijection, over-segmentation,...) - ###
    # - Create a bipartite graph where nodes represent the A labels (left) and B labels (right)
    #   and the edges the associations obtained using the max/min methods

    # - Reindex the labels
    target_labels = list(set(df_target.target.values) | set(df_reference.target.values))
    reference_labels = list(set(df_target.reference.values) | set(df_reference.reference.values))

    label_tp_list = [(m, 'l') for m in target_labels] + [(d, 'r') for d in reference_labels]
    lg2nid = dict(zip(label_tp_list, range(len(label_tp_list))))

    # - Create the graph
    G = nx.Graph()

    G.add_nodes_from([(nid, {'label': lab, 'group': g}) for (lab, g), nid in lg2nid.items()])

    target_to_ref_list = [(lg2nid[(i, 'l')], lg2nid[(j, 'r')]) for i, j in target_in_reference.items()]
    G.add_edges_from(target_to_ref_list)

    ref_to_target_list = [(lg2nid[(i, 'r')], lg2nid[(j, 'l')]) for i, j in reference_in_target.items()]
    G.add_edges_from(ref_to_target_list)

    # - Overlap analysis
    # - Get the  target_cells <--> reference_cells from the connected subgraph in G
    connected_graph = [list(G.subgraph(c)) for c in nx.connected_components(G)]

    # - Gather all the connected subgraph and reindex according to the image labels
    nid2lg = {v: k for k, v in lg2nid.items()}

    out_results = []
    for c in connected_graph:
        if len(c) > 1:  # at least two labels
            target, reference = [], []
            for nid in c:
                if nid2lg[nid][1] == 'l':
                    target.append(nid2lg[nid][0])  # label from target image
                else:
                    reference.append(nid2lg[nid][0])  # label from reference image

            out_results.append({'target': target, 'reference': reference})

    # - Add the background associations
    # - target --> reference background
    for lab in target_background:
        if lab != BACKGROUND_LABEL:  # ignore target background
            out_results.append({'target': [lab], 'reference': []})

    # - reference --> target background
    for lab in reference_background:
        if lab != BACKGROUND_LABEL:  # ignore reference background
            out_results.append({'target': [], 'reference': [lab]})

    out_results = pd.DataFrame(out_results)
        
    out_results['segmentation_state'] = out_results.apply(segmentation_state, axis=1) # add name for each type of association

    ignore_background = True

    cell_statistics = {'one-to-one': 0, 'over-segmentation': 0, 'under-segmentation': 0, 'misc.': 0}

    if not ignore_background:
        cell_statistics['background'] = 0 # add background

    state_target = {lab: state for list_lab, state
                    in zip(out_results.target.values, out_results.segmentation_state.values)
                    for lab in list_lab if state in cell_statistics}

    for lab, state in state_target.items():
        cell_statistics[state] += 1

    total_cells = len(state_target)
    cell_statistics = {state: np.around(val / total_cells * 100, 2) for state, val in cell_statistics.items()}

    # - add the missing  cells : percentage of reference cells that are missing in the predicted segmentation
    total_reference_cells = len(np.unique([item for sublist in out_results.reference.values for item in sublist]))
    missing_cells = [item for sublist in out_results.reference.loc[out_results.segmentation_state == 'missing'].values for item in sublist]

    total_missing = len(missing_cells)
    cell_statistics['missing'] = np.around(total_missing/total_reference_cells * 100, 2)

    stats_dict = dict (
        number_of_cells = total_cells,
        correct_segmentations = cell_statistics['one-to-one'],
        oversegmentation_rate = cell_statistics['over-segmentation'],
        undersegmentation_rate = cell_statistics['under-segmentation'],
        missing_rate = cell_statistics['missing']
    )

    return stats_dict
    
